#include <config.h>

#include "variable.h"
#include "structure.h"
#include "listiterator.h"
#include "error.h"

namespace dna
{

Variable::Variable(BLData* data)
{
    this->data = data;
}

Variable::~Variable()
{
}

const char* Variable::getName() const
{
    return data->var_name;
}

int Variable::getLength() const
{
    return data->length;
}

static const char* getValueType(BLData* data)
{
    switch (data->type) {
	case BCHR: return data->length > 1 ? "char[]" : "char";
	case BSHORT: return data->length > 1 ? "short[]" : "short";
	case BINT: return data->length > 1 ? "int[]" : "int";
	case BFLOAT: return data->length > 1 ? "float[]" : "float";
	case BSTRC: return data->length > 1 ? "structure[]" : "structure";
	case BPTR: return data->length > 1 ? "ptr[]" : "ptr";
	default: return "unknown";
    }
}

const char* Variable::getValueInternalString() const
{
    if (data->type != BPTR)
	throw Error("tried reading variable with type %s expected ptr",
		    getValueType(data));
    
    return (const char*) data->stringPtr;
}

const char* Variable::getValueString() const
{
    // check type
    if (data->type != BCHR)
	throw Error("tried reading variable with type %s expected char[]",
		    getValueType(data));

    return data->data.chr;
}

char Variable::getValueChar(int num) const
{
    if (num >= data->length)
	throw Error("Index out of range at variable '%s' (%d/%d)",
		    getName(), num, data->length);

    if (data->type != BCHR)
	throw Error("tried reading variable with type %s expected char",
		    getValueType(data));

    return data->data.chr[num];
}

short Variable::getValueShort(int num) const
{
    if (num >= data->length)
	throw Error("Index out of range at variable '%s' (%d/%d)",
		    getName(), num, data->length);
    
    if (data->type != BSHORT)
	throw Error("tried reading variable with type %s expected short",
		    getValueType(data));

    return data->data.shrt[num];
}

unsigned short Variable::getValueUShort(int num) const
{
    if (num >= data->length)
	throw Error("Index out of range at variable '%s' (%d/%d)",
		    getName(), num, data->length);
    
    if (data->type != BSHORT)
	throw Error("tried reading variable with type %s expected short",
		    getValueType(data));

    return ((unsigned short) data->data.shrt[num]);
}


int Variable::getValueInt(int num) const
{
    if (num >= data->length)
	throw Error("Index out of range at variable '%s' (%d/%d)",
		    getName(), num, data->length);
    
    if(data->type == BINT)
        return data->data.intg[num];
    else if(data->type == BSHORT)
        return data->data.shrt[num];
    else {
    	throw Error("tried reading variable with type %s expected int",
		    getValueType(data));
    }
}

float Variable::getValueFloat(int num) const
{
    if (num >= data->length)
	throw Error("Index out of range at variable '%s' (%d/%d)",
		    getName(), num, data->length);
    
    if (data->type != BFLOAT)
	throw Error("tried reading variable with type %s expected float",
		    getValueType(data));

    return data->data.flt[num];
}

Structure Variable::getValueStructure(int num) const
{
    if (num >= data->length)
	throw Error("Index out of range at variable '%s' (%d/%d)",
		    getName(), num, data->length);
    
    if (data->type == BPTR)
	return Structure( (BLStruct*) 
		BReadUINT32((char*) (data->data.ptr)));

    if (data->type != BSTRC)
	throw Error(
	    "tried reading variable with type %s expected structure or ptr",
		    getValueType(data));

    return Structure(data->data.strc + num);
}

ListIterator Variable::getValueListIterator(int num) const
{
    Structure base = getValueStructure(num);
    return ListIterator(base);
}

} // end of namespace dna

