#include <config.h>

#include <iostream>
#include <iomanip>
#include <fstream>

#include <sys/types.h>

#include "version.h"
#include "options.h"
#include <getopt/getopt.h>
#include <dna++/blendparser.h>
#include <dna++/error.h>
#include <writecs/xmlwriter.h>
#include <writecs/blend2cs.h>

Config config;

void usage(const char* appname)
{
    std::cerr << 
	"usage: " << appname << " [option] blenderfile \n"
    << "Converts a blender file into a Crystal Space file\n\n"
    << "Options:\n"
    << "  -h, --help		print this help message\n"
    << "  -v, --version		print version number of " << appname << "\n"
    << "  -l, --library		output as a CrystalSpace library file\n"
    << "  -o, --output          output file name, optional\n"
    << "  -p, --polygonnames    give polygons names (mainly for debugging)\n"
    << "  -w, --world		output as a CrystalSpace world or level (default)\n"
    << "  -t, --texturedir	path for textures\n"
    << "  -T, --terraindir	path for terrain stuff\n"
    << "  -f, --fixlights       convert unsupported lighttypes to default one\n"
    << "  -L, --listfile        dump blender file to output (don't convert)"
    // -L is for debugging only, so don't mention it here
    << std::endl;
}

void parseArguments (int argc, char** argv)
{
    bool userspecoutput=false;
    while (1)
    {
	static struct blend2cs::option long_options[] =
	{
	    {"help", blend2cs::no_argument, 0, 'h'},
	    {"version", blend2cs::no_argument, 0, 'v'},
	    {"library", blend2cs::no_argument, 0, 'l'},
            {"polygonnames", blend2cs::no_argument, 0, 'p'},
	    {"world", blend2cs::no_argument, 0, 'w'},
	    {"output", blend2cs::required_argument, 0, 'o'},
	    {"texturedir", blend2cs::required_argument, 0, 't'},
	    {"terraindir", blend2cs::required_argument, 0, 'T'},
	    {"listfile", blend2cs::no_argument, 0, 'L'},
	    {"fixlights", blend2cs::no_argument, 0, 'f'},
	    {0, 0, 0, 0}
	};

	int option_index = 0;
	int c = blend2cs::getopt_long
		(argc, argv, "hvlpwo:t:Lfd", long_options, &option_index);

	if (c == -1)
	    break;

	switch (c)
	{
	case 'v':
    	    printf ("blend2cs version %s\n", blend2cs::BLEND2CS_VERSION);
	    exit(0);
		
	case 'l':
	    config.outputlevel = false;
	    break;

        case 'p':
            config.polygonnames = true;
            break;

	case 'w':
	    config.outputlevel = true;
    	    break;

	case 'o':           // Sets the output file name.
	    config.outputfile = blend2cs::optarg;
	    userspecoutput=true;
	    break;

	case 't':
	    config.texturedir = blend2cs::optarg;
	    if (config.texturedir[config.texturedir.size()-1] != '/')
		config.texturedir.append("/");
	    break;
	    
	case 'T':
	    config.terraindir = blend2cs::optarg;
	    if (config.terraindir[config.terraindir.size()-1] != '/')
		config.terraindir.append("/");
	    break;
	    
	case 'h':
	    usage(argv[0]);
	    exit(0);

	case 'L':
	    config.printstructure = true;
	    break;
      
	case 'f':
	    config.fixlights = true;
	    break;
      
	default:
	    // getopt already reported the error
	    usage(argv[0]);
	    exit(1);
	}
    }

    if (argc - blend2cs::optind != 1)
    {
	usage(argv[0]);
	exit(1);
    }
    
    config.blenderfile = argv[blend2cs::optind++];
    if (!userspecoutput) {   // We have to generate the output file name
      config.outputfile=config.blenderfile;
      std::string::size_type pos=config.outputfile.rfind(".blend",
						config.outputfile.length() );
      if (pos==std::string::npos || pos<1) {
	std::cerr << "Your specified blender file name does not contain a\n"
		  << ".blend ending.  Please specify output file name with\n"
		  << "the --output option.\n";
	usage(argv[0]);
	exit(1);
      }
      std::string::size_type ltr=config.outputfile.length()-pos;
      std::string replc="";
      if (!config.outputlevel) replc=".lib";
      else replc=".world";
      config.outputfile.replace(pos,ltr,replc);  // Chop off ".blend".
    }
    //    config.outputfile = argv[blend2cs::optind++];
}

int main(int argc, char *argv[])
{
    using dna::BlendParser;
    using dna::Error;
    using blend2cs::Blend2CS;
    using blend2cs::XmlWriter;

    parseArguments (argc, argv);

    BlendParser* parser;
    try {
	parser = new BlendParser(config.blenderfile.c_str());
    } catch (Error& e) {
	std::cerr << argv[0] << ": " << e.what() << std::endl;
	exit(1);
    }
    if (config.printstructure) {
	parser->printStructure();
	exit(0);
    }

    std::ofstream output(config.outputfile.c_str());
    if(!output.good()) {
	std::cerr << "fatal: unable to open output file " << 
	    config.outputfile.c_str() << std::endl;
	exit(-1);
    }
    XmlWriter xml(output);

    Blend2CS converter(*parser, xml);
    converter.setTextureDir(config.texturedir);
    converter.setTerrainDir(config.terraindir);
#ifndef DEBUG
    try {
#endif
	converter.scan();
#ifndef DEBUG
    } catch (Error& e) {
	std::cerr << argv[0] << ": Error while scanning '"
	    << config.blenderfile << "':" << std::endl;
	std::cerr << "   " << e.what() << std::endl;
	exit(-1);
    }
#endif

    try {
	if (config.outputlevel == true)
	    converter.writeWorld();
	else
	    converter.writeLibrary();
    } catch (Error& e) {
	std::cerr << argv[0] << ": Error while outputting '" 
	    << config.outputfile << "':" << std::endl;
	std::cerr << "   " << e.what() << std::endl;
	exit(-1);
    }
  
    delete parser;
    return 0;
}
