/** This file writes a CrystalSpace map file from a parsed blender file. (Look
 * at blend2cs.h for details)
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <config.h>

#include <math.h>
#include <dna++/blendparser.h>
#include <dna++/variable.h>
#include <dna++/listiterator.h>
#include <dna++/blendertypes.h>

#include "xmlwriter.h"
#include "vector3d.h"
#include "matrix3d.h"
#include "object.h"
#include "sector.h"
#include "mesh.h"
#include "trianglemesh.h"
#include "blend2cs.h"
#include "thingmesh.h"
#include "terrainmesh.h"

namespace blend2cs
{

using dna::Variable;
using dna::Structure;

Blend2CS::Blend2CS(dna::BlendParser& newblend, XmlWriter& newxml)
    : blend(newblend), xml(newxml), worldsettings(0)
{
    usenewrenderer = false;
}

Blend2CS::~Blend2CS()
{
    SectorIterator i = getSectorIterator();
    while(i.hasNext())
	delete i.next();

    MeshIterator m = getMeshIterator();
    while(m.hasNext())
	delete m.next();

    delete worldsettings;
}

//---------------------------------------------------------------------------
// Phase 1

/** do the phase1 object scan (see README for details) */
void Blend2CS::scan()
{
    // Scan all scenes
    Structure top = blend.getTopLevelStructure();
    Structure::Iterator i = top.getIterator("Scene");
    while (i.hasNext()) {
	Structure scene = i.next();
	scanObjects(scene);
    }
}

void Blend2CS::scanObjects(Structure scene)
{
    std::string scenename = scene.getVariable("id.name").getValueString()+2;

    // Iterate over all objects in the scene
    dna::ListIterator i = scene.getVariable("base.first").getValueListIterator();
    while (i.hasNext()) {
	Structure base = i.next();
	Structure structure = base.getVariable("object").getValueStructure();

	Sector& sector = createGetSector(scenename);
	Object* object = sector.getObject(structure);
	// already parsed?
	if (object)
	    continue;

	// parse Object
	object = new Object(structure);
	if (!scanObject(sector, object))
	    delete object;
    }
}

/* returns true if the reference to the object has been consumed (ie. no delete
 * needed)
 */
bool Blend2CS::scanObject(Sector& sector, Object* object)
{
    /* check if we have an object with a special meaning like the WorldSettings
     * object. */
    if (checkSpecialObject(sector, object))
	return false;

    // Check for mesh types in hirarchical meshes
    Structure parent
	= object->getStructure().getVariable("parent").getValueStructure();
    if (parent.isGood() || object->hasChilds()) {
	if (object->getType() == Object::LIGHT) {
	    std::cerr << "Skipping Lamp '" << object->getName()
		<< "': lamps as parent/child objects not supported yet."
		<< std::endl;
	    return false;
	}
	if (object->getType() == Object::CAMERA) {
	    std::cerr << "Skipping Camera '" << object->getName()
		<< "': cameras as parent/child objects not supported yet."
		<< std::endl;
	    return false;
	}
    }

    // Create hirarchy
    switch(object->getType()) {
	case Object::THING:
	case Object::TERRAIN:
	case Object::NULLMESH:
	case Object::NODE:
	case Object::GENMESH:
	case Object::MESHREF:
	case Object::MESHLIB:
	case Object::SPRITE3D:
	    if (parent.isGood()) {
		Object* parentobj = sector.getObject(parent);
		if (!parentobj) {
		    parentobj = new Object(parent);
		    sector.addObject(parentobj);
		}
		parentobj->addChild(object);
	    }
	    break;
	default:
	    break;
    }

    // Parse Objects
    Mesh* mesh = 0;
    switch (object->getType()) {
	case Object::GENMESH:
	case Object::SPRITE3D:
	    mesh = scanTriangleMeshObject(object);
	    break;
	case Object::THING:
        case Object::PORTAL:
	    mesh = scanMeshObject(object);
	    break;
	case Object::TERRAIN:
	    mesh = scanTerrainMeshObject(object);
	    break;
	case Object::NULLMESH:
	case Object::CAMERA:
	case Object::NODE:
	case Object::LIGHT:
	case Object::MESHREF:
	case Object::MESHLIB:
	    break; // these objects are just fine
	default:
	    std::cerr << "Skipping object '" << object->getName()
		      << "' in scanObject: unsupported type." << std::endl;
	    break;
    }
    if(mesh) {
        // if the mesh is used more than once, then we should write out a
        // meshfact (if this is possible)
        if (mesh->getUseCount() == 1 && mesh->supportFactory())
            meshfacts.push_back(mesh);
        mesh->incUseCount();
    }

    sector.addObject(object);
    return true;
}

bool Blend2CS::checkSpecialObject(Sector& sector, Object* object)
{
    // Check for special world- and sectorsettings objects
    if (object->getName() == "WorldSettings") {
	if (worldsettings != 0) {
	    std::cerr << "Warning: multiple worldsettings specified. "
		<< "Overriding first one." << std::endl;
	}
	delete worldsettings;
	worldsettings = new WorldSettings(object->getStructure());

	return true;
    }
    if (object->getName().find("SectorSettings") != std::string::npos) {
	if (sector.getSettings() != 0) {
	    std::cerr << "Warning: multiple definitions of SectorSettings"
		<< " in Scene '" << sector.getName() << "'" << std::endl;
	    return true;
	}
	SectorSettings* sectorsettings
	    = new SectorSettings(object->getStructure());
	sector.setSettings(sectorsettings);

	return true;
    }

    return false;
}

Mesh* Blend2CS::scanTerrainMeshObject(Object* object)
{
    dna::Structure meshstructure
	= object->getStructure().getVariable("data").getValueStructure();


    if (!strcmp(meshstructure.getName(),"Mesh")) {   //This is a mesh structure
        // try to find an already existing mesh
        Mesh* mesh = findMesh(meshstructure, object->getMeshFlags());
        if (mesh)
            return mesh;

        TerrainMesh* terrainmesh
            = new TerrainMesh(meshstructure, object->getMeshFlags());

	addMesh(terrainmesh);
	meshfacts.push_back(terrainmesh);
        return terrainmesh;
    }

    // Uh, oh. We don't know what it is
    throw dna::Error("Object has data of an unknown type %s. Please "
            "contact blend2cs developers.",  meshstructure.getName());
}

Mesh* Blend2CS::scanMeshObject(Object* object)
{
    dna::Structure meshstructure
	= object->getStructure().getVariable("data").getValueStructure();


    if (!strcmp(meshstructure.getName(),"Mesh")) {   //This is a mesh structure
        // try to find an already existing mesh
        Mesh* mesh = findMesh(meshstructure, object->getMeshFlags());
        if (mesh)
            return mesh;

        ThingMesh* thingmesh
            = new ThingMesh(meshstructure, object->getMeshFlags(),
	          false, true, object->getType () == Object::PORTAL);

        // special handling if the mesh has no vertices or faces
	if (thingmesh->getVertexCount() == 0 ||
                thingmesh->getPolygonCount() == 0) {
            object->setType(Object::NULLMESH);
            return 0;
        }

	addMesh(thingmesh);
        return thingmesh;
    }

    if (!strcmp(meshstructure.getName(),"ID")) {
        // This is an ID structure that might reference a library:
        if (meshstructure.getVariable("lib").getValueStructure().isGood()) {
            //This object mesh is described in another blender file.
            // Add a dummy mesh object with multiple use counts, so it should be
            // written from a factory, but do not put it in meshfacts.
            addMesh(new ThingMesh(meshstructure,object->getMeshFlags(), true));
            return 0;

            // In the future we should add an option to follow the link to the
            // other blender file, and get the data, then write that out as
            // a mesh...
        }
    }

    // Uh, oh. We don't know what it is
    throw dna::Error("Object has data of an unknown type %s. Please "
            "contact blend2cs developers.",  meshstructure.getName());
}

Mesh* Blend2CS::scanTriangleMeshObject(const Object* object)
{
    dna::Structure meshstructure
	= object->getStructure().getVariable("data").getValueStructure();
    // already exists?
    TriangleMesh* mesh = dynamic_cast<TriangleMesh*> (
        findMesh(meshstructure, object->getMeshFlags()));
    if (mesh)
        return mesh;

    // here we're doing a stupid 2 phase parsing... this could be
    // optimized....
    ThingMesh* mmesh = new ThingMesh(meshstructure, object->getMeshFlags(),
            false, false);
    mesh = new TriangleMesh(*mmesh);
    delete mmesh;
    addMesh(mesh);
    meshfacts.push_back(mesh);

    return mesh;
}

Sector& Blend2CS::createGetSector(const std::string& name)
{
    std::map<std::string, Sector*>::iterator i
	= sectors.find(name);

    if (i == sectors.end()) {
	Sector* sector = new Sector(name.c_str());
	sectors.insert (std::make_pair(name, sector));
	return *sector;
    }

    return * ((*i).second);
}

Blend2CS::SectorIterator Blend2CS::getSectorIterator() const
{
    return SectorIterator(sectors);
}

//---------------------------------------------------------------------------
// Phase 2

void Blend2CS::advertise()
{
  std::stringstream cmnt;
  cmnt << "This file was generated from " << config.blenderfile
      << " with " PACKAGE_STRING;
  xml.comment(cmnt.str());
  xml.comment("http://projects.blender.org/projects/blend2cs");
}

void Blend2CS::writeRenderloop(const std::string& renderloop)
{
    if (renderloop == "std_rloop_terrainfixed") {
    	xml.openSection("addon");
	xml.writeTag("plugin");
	xml.write("crystalspace.renderloop.loop.loader");
	xml.writeTag("paramsfile");
	xml.write("/shader/std_rloop_terrainfixed.xml");
    	xml.closeSection("addon");

	xml.openSection("shaders");
	xml.openSection("shader");
	xml.writeTag("file");
	xml.write("/shader/terrain_fixed_base.xml");
	xml.closeSection("shader");
	xml.openSection("shader");
	xml.writeTag("file");
	xml.write("/shader/terrain_fixed_splatting.xml");
	xml.closeSection("shader");
	xml.closeSection("shaders");
    } else if (renderloop == "std_rloop_diffuse") {
    	xml.openSection("addon");
	xml.writeTag("plugin");
	xml.write("crystalspace.renderloop.loop.loader");
	xml.writeTag("paramsfile");
	xml.write("/shader/std_rloop_diffuse.xml");
    	xml.closeSection("addon");

	xml.openSection("shaders");
	xml.openSection("shader");
	xml.writeTag("file");
	xml.write("/shader/ambient.xml");
	xml.closeSection("shader");
	xml.openSection("shader");
	xml.writeTag("file");
	xml.write("/shader/light_bumpmap.xml");
	xml.closeSection("shader");
	xml.openSection("shader");
	xml.writeTag("file");
	xml.write("/shader/light.xml");
	xml.closeSection("shader");
	xml.closeSection("shaders");
    } else if (renderloop == "std_rloop_shadowed") {
    	xml.openSection("addon");
	xml.writeTag("plugin");
	xml.write("crystalspace.renderloop.loop.loader");
	xml.writeTag("paramsfile");
	xml.write("/shader/std_rloop_shadowed.xml");
    	xml.closeSection("addon");

	xml.openSection("shaders");
	xml.openSection("shader");
	xml.writeTag("file");
	xml.write("/shader/ambient.xml");
	xml.closeSection("shader");
	xml.openSection("shader");
	xml.writeTag("file");
	xml.write("/shader/light_bumpmap.xml");
	xml.closeSection("shader");
	xml.openSection("shader");
	xml.writeTag("file");
	xml.write("/shader/light.xml");
	xml.closeSection("shader");
	xml.closeSection("shaders");
    } else if (renderloop == "standard") {
    } else {
	std::cerr << "Invalid renderloop '" << renderloop << "'!" << std::endl;
    }
}

void Blend2CS::writeWorld()
{
    advertise();
    xml.openSection("world");

    writePlugins();
    if (worldsettings) {
        writeRenderloop(worldsettings->getRenderLoop());
    }
    writeTexturesMaterials();
    writeSettings();
    writeStarts();
    writeTerrainAddons();
    writeMeshFactories(false);
    writeSectors();

    xml.closeSection("world");
}

void Blend2CS::writeLibrary()
{
    advertise();
    xml.openSection("library");

    writePlugins();
    writeTexturesMaterials();
    writeMeshFactories(true);

    xml.closeSection("library");
}


void Blend2CS::writePlugins()
{
    // a default plugins section, we can make this code smarter later
    xml.openSection("plugins");
    if (!worldsettings || worldsettings->getThingStatic() == true) {
        xml.openSection("plugin");
        xml.writeAttribute("name", "thing");
	xml.writeTag("id");
        xml.write("crystalspace.mesh.loader.thing");
        xml.openSection("defaults");
	xml.writeTag("staticshape");
	xml.writeTag("staticpos");
        xml.closeSection("defaults");
        xml.closeSection("plugin");
    } else {
        xml.writeTag("plugin");
        xml.writeAttribute("name", "thing");
        xml.write("crystalspace.mesh.loader.thing");
    }
    xml.writeTag("plugin");
    xml.writeAttribute("name", "null");
    xml.write("crystalspace.mesh.loader.null");
    if (!worldsettings || worldsettings->getGenmeshStatic() == true) {
        xml.openSection("plugin");
        xml.writeAttribute("name", "genmesh");
	xml.writeTag("id");
        xml.write("crystalspace.mesh.loader.genmesh");
        xml.openSection("defaults");
	xml.writeTag("staticshape");
	xml.writeTag("staticpos");
        xml.closeSection("defaults");
        xml.closeSection("plugin");
    } else {
    	xml.writeTag("plugin");
    	xml.writeAttribute("name", "genmesh");
    	xml.write("crystalspace.mesh.loader.genmesh");
    }
    xml.writeTag("plugin");
    xml.writeAttribute("name", "sprite3d");
    xml.write("crystalspace.mesh.loader.sprite.3d");

    if (!worldsettings || worldsettings->getThingStatic() == true) {
        xml.openSection("plugin");
        xml.writeAttribute("name", "thingfact");
	xml.writeTag("id");
        xml.write("crystalspace.mesh.loader.factory.thing");
        xml.openSection("defaults");
	xml.writeTag("staticshape");
        xml.closeSection("defaults");
        xml.closeSection("plugin");
    } else {
    	xml.writeTag("plugin");
    	xml.writeAttribute("name", "thingfact");
    	xml.write("crystalspace.mesh.loader.factory.thing");
    }
    if (!worldsettings || worldsettings->getGenmeshStatic() == true) {
        xml.openSection("plugin");
        xml.writeAttribute("name", "genmeshfact");
	xml.writeTag("id");
        xml.write("crystalspace.mesh.loader.factory.genmesh");
        xml.openSection("defaults");
	xml.writeTag("staticshape");
        xml.closeSection("defaults");
        xml.closeSection("plugin");
    } else {
    	xml.writeTag("plugin");
    	xml.writeAttribute("name", "genmeshfact");
    	xml.write("crystalspace.mesh.loader.factory.genmesh");
    }

    // @@@ Only output if map contains terrain?
    xml.writeTag("plugin");
    xml.writeAttribute("name", "terrain");
    xml.write("crystalspace.mesh.loader.terrain");
    xml.writeTag("plugin");
    xml.writeAttribute("name", "terrainfact");
    xml.write("crystalspace.mesh.loader.factory.terrain");

    xml.closeSection("plugins");
}

void Blend2CS::writeTerrainBasematerials(
    std::vector<std::string>& textures,
    std::set<std::string>& basematerialnames,
    std::set<std::string>& splatmaterialnames)
{
    std::string terrainpath = "";
    if (worldsettings)
        terrainpath = worldsettings->getTerrainPath();
    std::vector<Mesh*>::iterator i;
    for (i=meshfacts.begin(); i!=meshfacts.end(); i++) {
    	const Mesh* mesh = *i;
	const TerrainMesh* terrainmesh = dynamic_cast<const TerrainMesh*>(mesh);
	if(terrainmesh) {
	    const TerrainMeshFlags& tflags = terrainmesh->getFlags().terrain_flags;
	    textures.push_back(tflags.basematerial_name);
	    xml.openSection("texture");
	    xml.writeAttribute("name", tflags.basematerial_name);
	    xml.writeTag("file");
	    xml.write(terrainpath + terraindir + tflags.basematerial);
	    xml.closeSection("texture");

	    basematerialnames.insert(tflags.basematerial_name);
	    std::vector<std::string> mp = terrainmesh->getMaterialPalette();
	    std::vector<std::string>::iterator ii = mp.begin();
	    for (; ii != mp.end(); ii++) {
	        std::string s = *ii;
	        splatmaterialnames.insert(s);
	    }
	}
    }
}

void Blend2CS::writeTexturesMaterials()
{
    Structure top = blend.getTopLevelStructure();
    std::vector<std::string> textures;

    std::string texturepath = "";
    if (worldsettings)
        texturepath = worldsettings->getTexturePath();

    xml.openSection("textures");
    // create a texture from every image present
    Structure::Iterator i = top.getIterator("Image");
    while(i.hasNext()) {

	Structure image = i.next();

	xml.openSection("texture");
        const char *txtrname=image.getVariable("id.name[24]").getValueString()+2;
	xml.writeAttribute("name", txtrname);
	textures.push_back(txtrname);

	xml.writeTag("file");
	std::string imagepath = processImagePath(
	    image.getVariable("name").getValueString());
	xml.write(texturepath + imagepath);
	xml.closeSection("texture");

	// If that image is packed into the .blend file, we try to dump it.
	// or we dump something. :)
	Structure pFile= image.findSubsequentStructure("PackedFile");
	if (pFile) {
	  const char *pfdata=pFile.getVariable("data").getValueStructure().getVariable("next").getValueInternalString();
	  // See if file by this name already exists
	  std::ifstream inpf(imagepath.c_str());
	  if (!inpf) {  // File does not already exists, or can't be read by us
	    std::ofstream outpf(imagepath.c_str());
	    if (outpf) {
	      outpf.write(pfdata, pFile.getVariable("size").getValueInt());
	      outpf.close();
	    }
	  } else inpf.close();
	} else {
	  // std::cerr << "No packed texture for " << txtrname << std::endl;
	}
    }

    // Set with base material names. These are the materials
    // that will require a base material shader.
    std::set<std::string> basematerialnames;
    // Set with splatter material names. These are the materials
    // that will require a splatter material shader.
    std::set<std::string> splatmaterialnames;
    writeTerrainBasematerials(textures, basematerialnames, splatmaterialnames);

    xml.closeSection("textures");

    xml.openSection("materials");
    // simplistic approach: each image becomes 1 texture and is transformed
    // into 1 cs material
    std::vector<std::string>::iterator tex;
    bool use_shaders = false;
    if (worldsettings && worldsettings->getRenderLoop() != "standard") {
	if (worldsettings->getRenderLoop() == "std_rloop_diffuse" ||
	    	worldsettings->getRenderLoop() == "std_rloop_shadowed")
	    use_shaders = true;
    }
    for (tex=textures.begin(); tex!=textures.end(); tex++) {
	if (splatmaterialnames.count(*tex)) {
	    xml.openSection("material");
	    xml.writeAttribute("name", *tex);
	    xml.writeTag("texture");
	    xml.write(*tex);
	    xml.closeSection("material");

	    xml.openSection("material");
	    xml.writeAttribute("name", "splat_"+*tex);
	    xml.writeTag("texture");
	    xml.write(*tex);

	    xml.writeTag("shadervar");
	    xml.writeAttribute("name", "texture scale");
	    xml.writeAttribute("type", "vector2");
	    xml.write("16,16");
	    xml.writeTag("shader");
	    xml.writeAttribute("type", "terrain splat");
	    xml.write("terrain_fixed_splatting");
	    xml.closeSection("material");
	} else {
	    xml.openSection("material");
	    xml.writeAttribute("name", *tex);
	    xml.writeTag("texture");
	    xml.write(*tex);
	    if (basematerialnames.count(*tex)) {
	        xml.writeTag("shader");
	        xml.writeAttribute("type", "ambient");
	        xml.write("terrain_fixed_base");
	    } else if (use_shaders) {
	        xml.writeTag("shader");
	        xml.writeAttribute("type", "ambient");
	        xml.write("ambient");
	        xml.writeTag("shader");
	        xml.writeAttribute("type", "diffuse");
	        xml.write("light");
	    }
	    xml.closeSection("material");
        }
    }
    xml.closeSection("materials");
}

void Blend2CS::writeSettings()
{
    xml.openSection("settings");

    if (!worldsettings || worldsettings->getClearZBuf() == true) {
	xml.writeTag("clearzbuf");
	xml.write("yes");
    }
    if (!worldsettings || worldsettings->getClearScreen() == true) {
    	xml.writeTag("clearscreen");
	xml.write("yes");
    }

    if (worldsettings && worldsettings->isAmbientSet()) {
	xml.writeTag("ambient");
	xml.writeAttribute("red", worldsettings->getAmbient().getX());
	xml.writeAttribute("green", worldsettings->getAmbient().getY());
	xml.writeAttribute("blue", worldsettings->getAmbient().getZ());
    }

    if (worldsettings && worldsettings->getRenderLoop() != "standard"
    		&& worldsettings->getGlobalRenderloop()) {
        xml.writeTag("renderloop");
	xml.write(worldsettings->getRenderLoop());
    }

    xml.closeSection("settings");
}

void Blend2CS::writeStarts()
{
    // Search for cameras and use them as start points
    SectorIterator i = getSectorIterator();
    while (i.hasNext())	{
	Sector* sector = i.next();

	Sector::ObjectIterator o = sector->getObjectIterator();
	while (o.hasNext()) {
	    Object* object = o.next();

	    if (object->getType() == Object::CAMERA)
		writeStart(sector, object);
	}
    }
}

void Blend2CS::writeStart(const Sector* sector, const Object* object)
{
    xml.openSection("start");
    xml.writeAttribute("name", object->getName());

    xml.writeTag("sector");
    xml.write(sector->getName());
    object->getPosition().write(xml, "position");

    // TODO: write forward vector

    xml.closeSection("start");
}

void Blend2CS::writeTerrainAddon(const TerrainMesh* terrainmesh)
{
    const TerrainMeshFlags& tflags = terrainmesh->getFlags().terrain_flags;
    std::string terrainpath = "";
    if (worldsettings)
        terrainpath = worldsettings->getTerrainPath();

    xml.openSection("addon");
    xml.writeAttribute("plugin", "crystalspace.terraformer.simple.loader");
    xml.writeTag("name");
    xml.write(terrainmesh->getName());
    xml.writeTag("heightmap");
    xml.write(terrainpath+terraindir+tflags.heightmap);
    const Box2D& bounds = terrainmesh->getBounds();
    float hmin = terrainmesh->getHeightMin();
    float hmax = terrainmesh->getHeightMax();
    float hdiff = hmax-hmin;
    xml.writeTag("scale");
    xml.writeAttribute("x", (bounds.getMax().getX()-bounds.getMin().getX())/2.0f);
    xml.writeAttribute("y", hdiff);
    xml.writeAttribute("z", (bounds.getMax().getY()-bounds.getMin().getY())/2.0f);
    xml.writeTag("offset");
    xml.writeAttribute("x", 0);
    xml.writeAttribute("y", hmin);
    xml.writeAttribute("z", 0);
    xml.closeSection("addon");
}

void Blend2CS::writeTerrainAddons()
{
    std::vector<Mesh*>::iterator i;
    for (i=meshfacts.begin(); i!=meshfacts.end(); i++) {
    	const Mesh* mesh = *i;
	const TerrainMesh* terrainmesh = dynamic_cast<const TerrainMesh*>(mesh);
	if(terrainmesh) {
	    writeTerrainAddon(terrainmesh);
	}
    }
}

void Blend2CS::writeMeshFactories(bool allmeshes)
{
    if (!allmeshes) {
	// only meshs in meshfacts vector
    	std::vector<Mesh*>::iterator i;
	for (i=meshfacts.begin(); i!=meshfacts.end(); i++) {
	    writeMeshFactory(*i);
	}
    } else {
	// all meshfactories (this is used in library output mode, because there
	// we even want a meshfactory for meshes which are only used once)
	MeshIterator i = getMeshIterator();
	while(i.hasNext()) {
	    const Mesh* mesh = i.next();

	    if (!mesh->supportFactory()) {
		std::cerr << "Skipping Mesh '" << mesh->getName() << "':"
		          << " moveable not supported for libraries/factories."
			  << std::endl;
		continue;
	    }
	    writeMeshFactory(mesh);
	}
    }
}

void Blend2CS::writeMeshFactory(const Mesh* mesh)
{
    xml.openSection("meshfact");
    xml.writeAttribute("name", mesh->getName());
    xml.writeTag("plugin");
    xml.write(mesh->getMeshFactoryName());

    mesh->write(xml);

    xml.closeSection("meshfact");
}

void Blend2CS::writeSectors()
{
    std::map<std::string, Sector*>::iterator i;

    for (i=sectors.begin(); i!=sectors.end(); i++) {
	Sector* sector = i->second;

	xml.openSection("sector");
	xml.writeAttribute("name", sector->getName());

	writeSector(sector);

	xml.closeSection("sector");
    }
}

void Blend2CS::writeSector(const Sector* sector)
{
    if (worldsettings && worldsettings->getRenderLoop() != "standard"
    		&& !worldsettings->getGlobalRenderloop()) {
        xml.writeTag("renderloop");
	xml.write(worldsettings->getRenderLoop());
    }

    Sector::ObjectIterator i = sector->getObjectIterator();

    while (i.hasNext()) {
	Object* object = i.next();

	writeObject(object);
    }

    if (!sector->getSettings())
	return;

    const SectorSettings* settings = sector->getSettings();
    if (settings->getFogDensity() != 0) {
	xml.writeTag("fog");
	xml.writeAttribute("red", settings->getFogColor().getX());
	xml.writeAttribute("green", settings->getFogColor().getY());
	xml.writeAttribute("blue", settings->getFogColor().getZ());
	xml.writeAttribute("density", settings->getFogDensity());
    }
    const std::string& culler = settings->getCuller();
    xml.writeTag("cullerp");
    xml.writeAttribute("plugin", culler);
}

void Blend2CS::writeObject(const Object* object)
{
    // skip child meshes here
    if (object->getParent())
	return;

    switch (object->getType()) {
	case Object::LIGHT:
	    writeLightObject(object);
	    break;
	case Object::NULLMESH:
	    writeNullMeshObject(object);
	    break;
        case Object::NODE:
            writeNodeObject(object);
            break;
	case Object::GENMESH:
	case Object::THING:
	case Object::TERRAIN:
	    writeMeshObject(object);
	    break;
        case Object::PORTAL:
            writePortal(object);
            break;
	case Object::MESHREF:
            writeMeshRef(object);
            break;
	case Object::MESHLIB:
            writeMeshLib(object);
            break;
	case Object::CAMERA:
	    // ignore cameras here
	    break;
	default:
    	    std::cerr << "Skipping unsupported object '"
		<< object->getName() << "'" << std::endl;
	    break;
    }
}

void Blend2CS::writeLightObject(const Object* object)
{
    Structure light = object->getData();

    short type = light.getVariable("type").getValueShort();
    short mode = light.getVariable("mode").getValueShort();
    if (type != 0) {
		if (config.fixlights) {
			std::cout << "Warning: Convert Light '"
                                  << object->getName()
                                  << "' to type 'normal light'." << std::endl;
		} else {
                    std::cerr << "Skipping Light '" << object->getName() <<
                        "': only type 'Lamp' with mode 'Blend' supported" <<
                        std::endl;
                    std::cerr << "(You might try the --fixlights flag)" <<
                        std::endl;
                }
	return;
    }
    if (mode != 1) {
		if (config.fixlights) {
			std::cout << "Fixing mode of light '"
                            << object->getName() << "'" << std::endl;
		} else {
	std::cerr << "Skipping Light '" << object->getName()
	    << "': only Lamps with mode shadows are supported. "
	    << "(and this is the only supported mode)" << std::endl
            << "(You might try the --fixlight flag)"
	    << std::endl;
	return;
		}
    }

    xml.openSection("light");

    xml.writeAttribute("name", object->getName());

    // Position
    object->getPosition().write(xml, "center");

    // Color
    xml.writeTag("color");
    xml.writeAttribute("red", light.getVariable("r").getValueFloat());
    xml.writeAttribute("green", light.getVariable("g").getValueFloat());
    xml.writeAttribute("blue", light.getVariable("b").getValueFloat());

    // Radius
    xml.writeTag("radius");
    if (usenewrenderer) {
	xml.writeAttribute("brightness",
	    light.getVariable("energy").getValueFloat());
    }
    xml.write(light.getVariable("dist").getValueFloat());

    writeLightProperties(object);
    writeProperties(object);

    xml.closeSection("light");
}

void Blend2CS::writeMeshObject(const Object* object)
{
    dna::Structure meshstructure
	= object->getStructure().getVariable("data").getValueStructure();
    const Mesh* mesh = findMesh(meshstructure, object->getMeshFlags());
    if (!mesh)
	throw dna::Error("internal Error at Blend2CS::writeMeshObject.");

    const MeshFlags& meshflags = mesh->getFlags ();
    if (!meshflags.exported) return;

    xml.openSection("meshobj");
    xml.writeAttribute("name", object->getName());

    xml.writeTag("plugin");
    xml.write(mesh->getMeshPluginName());

    bool usefactory = mesh->supportFactory() && mesh->getUseCount() > 1;
    if (usefactory) {
	// we have a meshfactory
	xml.openSection("params");
	xml.writeTag("factory");
	xml.write(mesh->getName());

        const TriangleMesh* trianglemesh = dynamic_cast<const TriangleMesh*>(mesh);
        if(trianglemesh) {
            xml.writeTag("material");
            if (trianglemesh->getMaterial() != "")
                xml.write(trianglemesh->getMaterial());
            else
                xml.write("no");
        }

	const TerrainMesh* terrainmesh = dynamic_cast<const TerrainMesh*>(mesh);
	if(terrainmesh) {
	    std::string terrainpath = "";
	    if (worldsettings)
	        terrainpath = worldsettings->getTerrainPath();
	    const TerrainMeshFlags& tflags = terrainmesh->getFlags().terrain_flags;
	    if (tflags.saveheightmap)
	        terrainmesh->saveHeightMap("");
	    if (tflags.savematerialmap)
	        terrainmesh->saveMaterialPalette("");
	    if (tflags.savebasematerial)
	        terrainmesh->saveBaseMaterial("");
	    xml.writeTag("material");
            xml.write(tflags.basematerial_name);

	    xml.openSection("materialpalette");
	    const std::vector<std::string> mp = terrainmesh->getMaterialPalette();
	    VectorIterator<std::string> it = VectorIterator<std::string> (mp);
	    while (it.hasNext()) {
	        xml.writeTag("material");
		std::string mat = it.next();
	        xml.write("splat_"+mat);
	    }
	    xml.closeSection("materialpalette");

	    xml.writeTag("lodvalue");
	    xml.writeAttribute("name", "splatting distance");
	    xml.write(tflags.splatting_distance);
	    xml.writeTag("lodvalue");
	    xml.writeAttribute("name", "block resolution");
	    xml.write(tflags.block_resolution);
	    xml.writeTag("lodvalue");
	    xml.writeAttribute("name", "block split distance");
	    xml.write(tflags.block_split_distance);
	    xml.writeTag("lodvalue");
	    xml.writeAttribute("name", "minimum block size");
	    xml.write(tflags.minimum_block_size);
	    xml.writeTag("lodvalue");
	    xml.writeAttribute("name", "cd resolution");
	    xml.write(tflags.cd_resolution);
	    xml.writeTag("lodvalue");
	    xml.writeAttribute("name", "cd lod cost");
	    xml.write(tflags.cd_lod_cost);
	    xml.writeTag("lodvalue");
	    xml.writeAttribute("name", "lightmap resolution");
	    xml.write(tflags.lightmap_resolution);

	    xml.writeTag("materialmap");
	    xml.writeAttribute("image", terrainpath+terraindir+tflags.materialmap);

	    xml.writeTag("staticlighting");
	    xml.write(tflags.staticlighting ? "true" : "false");
	    xml.writeTag("castshadows");
	    xml.write(tflags.castshadows ? "true" : "false");
	}

	xml.closeSection("params");
    } else {
	// we write the meshobject normally
        mesh->write(xml);
    }

    // write childs
    Object::ChildIterator i = object->getChildIterator();
    while(i.hasNext()) {
	Object* child = i.next();
	writeMeshObject(child);
    }

    // Disabling collision detection or visibility culling.
    if ((!meshflags.colldet) || (!meshflags.viscull)) {
        xml.openSection("polymesh");
	if (!meshflags.colldet) {
	    xml.writeTag("colldet");
	}
	if (!meshflags.viscull) {
	    xml.writeTag("viscull");
	}
	xml.closeSection("polymesh");
    }

    if (meshflags.noclip) {
        xml.writeTag("noclip");
    }

    writeCloseMesh(object, "meshobj", true);
}

void Blend2CS::writeCloseMesh(const Object* object, const char* sectionName,
		bool move)
{
    switch (object->getZMode()) {
	case Object::ZUSE: xml.writeTag("zuse"); break;
	case Object::ZFILL: xml.writeTag("zfill"); break;
	case Object::ZTEST: xml.writeTag("ztest"); break;
	case Object::ZNONE: xml.writeTag("znone"); break;
	default: throw dna::Error("unknown zmode at Blend2CS::writeCloseMesh");
    }
    Object::FlagIterator f = object->getFlagIterator();
    while(f.hasNext()) {
	Object::Flag* flag = f.next();
	xml.writeTag(flag->name.c_str());
	if (flag->value != "") {
	    xml.write(flag->value);
	}
    }

    writeTransformation(object, move);
    writeProperties(object);

    xml.closeSection(sectionName);
}

void Blend2CS::writeMeshRef(const Object* object)
{
    xml.openSection("meshref");
    xml.writeAttribute("name", object->getName());
    writeCloseMesh(object, "meshref", true);
}

void Blend2CS::writeMeshLib(const Object* object)
{
    xml.openSection("meshlib");
    xml.writeAttribute("name", object->getName());
    writeCloseMesh(object, "meshlib", true);
}

void Blend2CS::writePortal(const Object* object)
{
    xml.openSection("portals");
    xml.writeAttribute("name", object->getName());

    dna::Structure meshstructure
	= object->getStructure().getVariable("data").getValueStructure();
    const ThingMesh* mesh = dynamic_cast<ThingMesh*> (
        findMesh(meshstructure, object->getMeshFlags()));
    if (!mesh)
	throw dna::Error("internal Error at Blend2CS::writePortal.");

    ThingMesh::PolygonIterator it = mesh->getPolygonIterator();
    while(it.hasNext()) {
        Polygon* polygon = it.next();

        xml.openSection("portal");

        for(int i = 0; i<polygon->getVertexCount(); ++i) {
            size_t vertexid = polygon->getVertex(i);
            mesh->getVertex(vertexid).write(xml, "v");
        }

        xml.writeTag("sector");
        xml.write(object->getPortalSector());

        int portalflags = object->getPortalFlags();
        if(portalflags & Object::PORTAL_ZFILL)
            xml.writeTag("zfill");
        if(portalflags & Object::PORTAL_FLOAT)
            xml.writeTag("float");
        if(portalflags & Object::PORTAL_CLIP)
            xml.writeTag("clip");
        if(portalflags & Object::PORTAL_MIRROR)
            xml.writeTag("mirror");
        if(portalflags & Object::PORTAL_STATIC)
            xml.writeTag("static");
        if(portalflags & Object::PORTAL_AUTORESOLVE)
            xml.writeTag("autoresolve");
        if(portalflags & Object::PORTAL_CLIPSTRADDLING)
            xml.writeTag("clipstraddling");
        if(portalflags & Object::PORTAL_VISCULL)
            xml.writeTag("viscull");
	//        if(polygon->getColldet())  xml.writeTag("colldet");

        if(object->getPortalMaxVisit() > 0) {
            xml.writeTag("maxvisit");
            xml.write(object->getPortalMaxVisit());
        }

        if(object->getPortalWarpBefore() != Vector3D(0,0,0)) {
            object->getPortalWarpBefore().write(xml, "wv");
        }
        if(object->getPortalWarpAfter() != Vector3D(0,0,0)) {
            object->getPortalWarpAfter().write(xml, "ww");
        }
        if(object->getPortalRotate() != Vector3D(0,0,0)) {
            xml.openSection("matrix");
            xml.writeTag("rotx");
            xml.write(object->getPortalRotate().getX());
            xml.writeTag("roty");
            xml.write(object->getPortalRotate().getY());
            xml.writeTag("rotz");
            xml.write(object->getPortalRotate().getZ());
            xml.closeSection("matrix");
        }

        xml.closeSection("portal");
    }

    writeCloseMesh(object, "portals", false);	// use hardmove
}

void Blend2CS::writeNodeObject(const Object* object)
{
  // This code writes out CS nodes.
    xml.openSection("node");
    xml.writeAttribute("name", object->getName());

    object->getPosition().write(xml,"position");

    Matrix3D m;
    Vector3D rot = object->getRotation();
    m.calcRotation(rot.getX(), rot.getZ(), rot.getY());
//printf ("'%s' rotation %g,%g,%g\n", object->getName().c_str(), rot.getX(),rot.getY(),rot.getZ());
    Vector3D vx (1, 0, 0);
    vx = m * vx;
    Vector3D vy (0, 0, 1);
    vy = m * vy;
    Vector3D vz (0, 1, 0);
    vz = m * vz;
    vx.write(xml, "xvector");
    vy.write(xml, "yvector");
    vz.write(xml, "zvector");
//m.dump();
//printf ("    %g,%g,%g\n", vx.getX(),vx.getY(),vx.getZ());
//printf ("    %g,%g,%g\n", vy.getX(),vy.getY(),vy.getZ());
//printf ("    %g,%g,%g\n", vz.getX(),vz.getY(),vz.getZ());

    writeProperties(object);
    xml.closeSection("node");
}

void Blend2CS::writeNullMeshObject(const Object* object)
{
    xml.openSection("meshobj");
    xml.writeAttribute("name", object->getName());

    xml.openSection("nullmesh");
    Vector3D(0,0,0).write(xml, "min");
    Vector3D(0,0,0).write(xml, "max");
    xml.closeSection("nullmesh");
    writeCloseMesh(object, "meshobj", true);
}

void Blend2CS::writeSprite3dObject(const Object* object)
{
    (void) object;
    // TODO
    return;

    xml.openSection("meshobj");

    xml.closeSection("meshobj");
}

void Blend2CS::writeTransformation(const Object* object, bool move)
{
    if (move)
	xml.openSection("move");
    else
	xml.openSection("hardmove");

    object->getPosition().write(xml, "v");

    if(object->getRotation() != Vector3D(0,0,0)) {
	xml.openSection("matrix");

	if (object->getRotation() != Vector3D(0,0,0)) {

	    // we have to use the yzx order to get the desired output
	    if (object->getRotation().getY() != 0) {
		xml.writeTag("roty");
		xml.write(object->getRotation().getY());
	    }
	    if (object->getRotation().getZ() != 0) {
		xml.writeTag("rotz");
		xml.write(object->getRotation().getZ());
	    }
	    if (object->getRotation().getX() != 0) {
		xml.writeTag("rotx");
		xml.write(object->getRotation().getX());
	    }
	}

	xml.closeSection("matrix");
    }

    if (move)
	xml.closeSection("move");
    else
	xml.closeSection("hardmove");
}

void Blend2CS::writeLightProperties(const Object* object)
{
    Object::PropertyIterator i = object->getPropertyIterator();

    while(i.hasNext()) {
	Property* property = i.next();
	const std::string& pname = property->getName();
	if (pname == "dynamic") {
	    xml.writeTag("dynamic");
	}
    }
}

void Blend2CS::writeProperties(const Object* object)
{
    Object::PropertyIterator i = object->getPropertyIterator();

    while(i.hasNext()) {
	Property* property = i.next();

	xml.writeTag("key");
	xml.writeAttribute("name", property->getName());
	xml.writeAttribute("value", property->getValue());
    }
}

/** transforms the path of a texturefile */
std::string Blend2CS::processImagePath(const std::string& path)
{
    const char* p = path.c_str();
    const char* filenamestart = p;
    for (; *p != 0; p++)
	if (*p == '/' || *p == '\\')
	    filenamestart = p+1;

    return texturedir + filenamestart;
}

void Blend2CS::setTextureDir(const std::string& newtexturedir)
{
    texturedir = newtexturedir;
}

void Blend2CS::setTerrainDir(const std::string& newterraindir)
{
    terraindir = newterraindir;
}

Blend2CS::MeshIterator Blend2CS::getMeshIterator() const
{
    return MeshIterator(meshes);
}

void Blend2CS::addMesh(Mesh* mesh)
{
    // Make sure the mesh has a unique name
    int nameadd = 0;
    std::string origname = mesh->getName();
    bool exists;
    do {
        exists = false;
        for(Meshes::const_iterator i = meshes.begin(); i != meshes.end(); ++i) {
            const Mesh* omesh = i->second;
            if(mesh->getName() == omesh->getName()) {
                exists = true;
                break;
            }
        }
        if(exists) {
            std::stringstream newname;
            newname << origname << '_' << (nameadd++);
            mesh->setName(newname.str());
        }
    } while(exists);

    meshes.insert(std::make_pair(mesh->getStructure(), mesh));
}

Mesh* Blend2CS::findMesh(dna::Structure structure, const MeshFlags& flags) const
{
    std::pair<Meshes::const_iterator, Meshes::const_iterator> range =
        meshes.equal_range(structure);
    for(Meshes::const_iterator i = range.first; i != range.second; ++i) {
        Mesh* mesh = i->second;
        if(mesh->getFlags() == flags)
            return i->second;
    }

    return 0;
}

} // end of namespace blend2cs
