#ifndef __BLEND2CS_BLENDCONV_H__
#define __BLEND2CS_BLENDCONV_H__

#include <string>
#include <map>
#include <set>
#include <dna++/structure.h>

#include "mapiterator.h"
#include "worldsettings.h"
#include "options.h"

namespace blend2cs
{

class BlendParser;
class XmlWriter;
class Mesh;
class Sector;
class Object;
class Vector3D;
class Matrix3D;
class MeshFlags;
class ThingMesh;
class TerrainMesh;

/**
 * This class converts a parsed Blender file into a CS xml format
 *
 * The parsing happens in 2 phases. In phase1 we'll construct a tree of
 * widgets and their childs in particular we construct a list of childs for
 * each object (blender only saves the parent of each object).
 * Also the separation into sectors is done in the phase 1 scan. As sectors
 * have to be written in order in CS, while the blender objects to sector
 * mapping depends on scenes, layers and/or override in properties. So we have
 * to construct a list of sectors first and write that out in order later.
 * In Phase2 we'll write out the CS map.
 */
class Blend2CS
{
public:
    Blend2CS(dna::BlendParser& blend, XmlWriter& xml);
    ~Blend2CS();

    /** sets directory for the textures */
    void setTextureDir(const std::string& dir);
    /** sets directory for the terrain stuff */
    void setTerrainDir(const std::string& dir);

    /** scan list of sectors/objects (Phase1) */
    void scan();
    
    /** write a CS worldfile (Phase2) */
    void writeWorld();
    /** write a CS library file */
    void writeLibrary();

private:
    void advertise();   
    /**
     * writes out base textures for terrain and push on textures
     * array.
     */
    void writeTerrainBasematerials(std::vector<std::string>& textures,
    	std::set<std::string>& basematerialnames,
    	std::set<std::string>& splatmaterialnames);
    /** writes texture and material definitions */
    void writeTexturesMaterials();
    /** writes world settings section */
    void writeSettings();
    /** writes plugins section */
    void writePlugins();
    /** writes starts section */
    void writeStarts();
    /** writes out terrain addon */
    void writeTerrainAddon(const TerrainMesh* mesh);
    /** writes out terrain addons */
    void writeTerrainAddons();
    /** writes out meshfactories. if allmeshes = false, then we'll only write
     * the ones in meshes vector, otherwise all meshs
     */
    void writeMeshFactories(bool allmeshs = false);
    /** writes a single meshfactory */
    void writeMeshFactory(const Mesh* mesh);
    void writeSectors();
    void writeRenderloop(const std::string& renderloop);

    void writeStart(const Sector* sector, const Object* object);
    void writeSector(const Sector* sector);

    /**
     * Write the closing of a mesh object, including zmode,
     * flags like render priority, the transformation, and the
     * final closing tag.
     */
    void writeCloseMesh(const Object* object, const char* sectionName, bool move);
    void writeObject(const Object* object);
    void writeLightObject(const Object* object);
    void writeMeshObject(const Object* object);
    void writeNodeObject(const Object* object);
    void writePortal(const Object* object);
    void writeNullMeshObject(const Object* object);
    void writeGenMeshObject(const Object* object);
    void writeSprite3dObject(const Object* object);
    void writeMeshRef(const Object* object);
    void writeMeshLib(const Object* object);

    void writeTransformation(const Object* object, bool move = false);
    void writeLightProperties(const Object* object);
    void writeProperties(const Object* object);

    /** scans all blenderobjects in a scene */
    void scanObjects(dna::Structure scene);
    bool scanObject(Sector& sector, Object* object);
    bool checkSpecialObject(Sector& sector, Object* object);
    /** scans a blender meshobject */
    Mesh* scanMeshObject(Object* object);
    Mesh* scanTriangleMeshObject(const Object* object);
    Mesh* scanTerrainMeshObject(Object* object);


    /** return an existing sector or create a new one */
    Sector& createGetSector(const std::string& name);
    
    typedef MapIterator<std::string, Sector*> SectorIterator;
    SectorIterator getSectorIterator() const;

    typedef MultiMapIterator<dna::Structure, Mesh*> MeshIterator;
    MeshIterator getMeshIterator() const;
    Mesh* findMesh(dna::Structure structure, const MeshFlags& flags) const;
    void addMesh(Mesh* mesh);

    std::string processImagePath(const std::string& path);
    
    std::string texturedir;
    std::string terraindir;
    
    dna::BlendParser& blend;
    XmlWriter& xml;
    /** produce map files which use features of the new CS OpenGL renderer in
     * developmont.
     */
    bool usenewrenderer;
    std::map<std::string, Sector*> sectors;

    typedef std::multimap<dna::Structure, Mesh*> Meshes;
    Meshes meshes;
    typedef std::vector<Mesh*> MeshFacts;
    MeshFacts meshfacts;

    WorldSettings* worldsettings;
};

} // end of namespace blend2cs

#endif

