#include <config.h>

#include <cmath>
#include <iostream>
#include <sstream>

#include <dna++/error.h>
#include <dna++/variable.h>
#include <dna++/listiterator.h>
#include <dna++/blendertypes.h>

#include "object.h"
#include "matrix4d.h"

namespace blend2cs
{

Object::Object(const dna::Structure& newstructure)
    : structure(newstructure), 
      portalflags(0), portalmaxvisit(0), zmode(ZUSE), parent(0)
{
    parseObject();
}

Object::~Object()
{
    PropertyIterator i = getPropertyIterator();
    while(i.hasNext())
	delete i.next();
    FlagIterator f = getFlagIterator();
    while (f.hasNext())
	delete f.next();
}

void Object::addChild(Object* object)
{
    childs.push_back(object);

    // change child
    object->parent = this;

    // Correct location for hierarhical transforms.
    dna::Variable parent_obmat = structure.getVariable("obmat");
    Matrix4D parent_m (
    	parent_obmat.getValueFloat (0),
    	parent_obmat.getValueFloat (1),
    	parent_obmat.getValueFloat (2),
    	parent_obmat.getValueFloat (3),
    	parent_obmat.getValueFloat (4),
    	parent_obmat.getValueFloat (5),
    	parent_obmat.getValueFloat (6),
    	parent_obmat.getValueFloat (7),
    	parent_obmat.getValueFloat (8),
    	parent_obmat.getValueFloat (9),
    	parent_obmat.getValueFloat (10),
    	parent_obmat.getValueFloat (11),
    	parent_obmat.getValueFloat (12),
    	parent_obmat.getValueFloat (13),
    	parent_obmat.getValueFloat (14),
    	parent_obmat.getValueFloat (15)
	);
    dna::Variable child_obmat = object->structure.getVariable("obmat");
    Matrix4D child_m (
    	child_obmat.getValueFloat (0),
    	child_obmat.getValueFloat (1),
    	child_obmat.getValueFloat (2),
    	child_obmat.getValueFloat (3),
    	child_obmat.getValueFloat (4),
    	child_obmat.getValueFloat (5),
    	child_obmat.getValueFloat (6),
    	child_obmat.getValueFloat (7),
    	child_obmat.getValueFloat (8),
    	child_obmat.getValueFloat (9),
    	child_obmat.getValueFloat (10),
    	child_obmat.getValueFloat (11),
    	child_obmat.getValueFloat (12),
    	child_obmat.getValueFloat (13),
    	child_obmat.getValueFloat (14),
    	child_obmat.getValueFloat (15)
	);
    parent_m.Invert ();
    Matrix4D result = child_m * parent_m;
    // @@@ This code only corrects the position of the object and not the rotation.
    object->position.set (result.Get(3,0),result.Get(3,2),result.Get(3,1));
}

Object::ChildIterator Object::getChildIterator() const
{
    return ChildIterator(childs);
}

Object::PropertyIterator Object::getPropertyIterator() const
{
    return PropertyIterator(properties);
}

const Property* Object::getProperty(const std::string& name) const
{
    PropertyIterator i = getPropertyIterator();
    while(i.hasNext()) {
	const Property* property = i.next();
	if (property->getName() == name)
	    return property;
    }
    
    return 0;
}

dna::Structure Object::getData() const
{
    return structure.getVariable("data").getValueStructure();
}

Object::FlagIterator Object::getFlagIterator() const
{
    return FlagIterator(flags);
}

static inline float radToDeg(float rad)
{
    return /*180*rad/M_PI*/ rad;
}

void Object::parseObject()
{
    name = structure.getVariable("id.name").getValueString() +2; //+2 skips OB

    switch(structure.getVariable("type").getValueShort()) {
	case OB_EMPTY: type = NODE; break;
	case OB_MESH: type = THING; break;
	case OB_LAMP: type = LIGHT; break;
	case OB_CAMERA: type = CAMERA; break;
	default:
	    std::cerr << "Skipping object '" << name
		<< "': unsupported type." << std::endl;
	    type = UNKNOWN;
	    break;
    }

    // parse transform, remember to swap y and z
    dna::Variable loc = structure.getVariable("loc");
    dna::Variable dloc = structure.getVariable("dloc");
    //dna::Variable orig = structure.getVariable("orig");
    position.set (loc.getValueFloat(0) + dloc.getValueFloat(0),// - orig.getValueFloat(0),
		  loc.getValueFloat(2) + dloc.getValueFloat(2),// - orig.getValueFloat(2),
		  loc.getValueFloat(1) + dloc.getValueFloat(1));// - orig.getValueFloat(1));

    dna::Variable rot = structure.getVariable("rot");
    dna::Variable drot = structure.getVariable("drot");
    rotation.set ( -(rot.getValueFloat(0) + drot.getValueFloat(0)),
		   (rot.getValueFloat(2) + drot.getValueFloat(2)),
		   -(rot.getValueFloat(1) + drot.getValueFloat(1)));

    dna::Variable size = structure.getVariable("size");
    dna::Variable dsize = structure.getVariable("dsize");
    meshflags.scale.set (size.getValueFloat(0) + dsize.getValueFloat(0),
	       size.getValueFloat(2) + dsize.getValueFloat(2),
	       size.getValueFloat(1) + dsize.getValueFloat(1));

    // parse Properties
    dna::ListIterator i 
	= structure.getVariable("prop.first").getValueListIterator();
    while(i.hasNext()) {
	dna::Structure propstructure = i.next();
	Property* property = new Property(propstructure);

	// try to interpret property and remember all properties we don't
	// interpret
	if (!interpretProperty(property))
    	    properties.push_back(property);
    }
}

void Object::parseBoolProperty(const Property* property, bool& var)
{
    try {
	var = property->getValueBool();
    } catch (dna::Error e) {
	std::cerr << "Couldn't parse property '" << property->getName() << 
	"': Couldn't parse boolean value '" << property->getValue() << "'"
	<< " in object '" << getName() << "'" << std::endl;
    }
}

bool Object::interpretProperty(const Property* property)
{
    const std::string& pname = property->getName();

    if (pname == "type") {
	if (type != THING) {
	    std::cerr << "type property only supported for meshes "
		<< " in object '" << getName() << "'" << std::endl;
	    return true;
	}
	if (property->getValue() == "null") {
	    type = NULLMESH;
	} else if (property->getValue() == "thing") {
	    type = THING;
	} else if (property->getValue() == "terrain") {
	    type = TERRAIN;
        } else if (property->getValue() == "portal") {
            type = PORTAL;
	} else if (property->getValue() == "genmesh") {
	    type = GENMESH;
	} else if (property->getValue() == "sprite3d") {
	    type = SPRITE3D;
	} else if (property->getValue() == "meshref") {
	    type = MESHREF;
	} else if (property->getValue() == "meshlib") {
	    type = MESHLIB;
	} else if (property->getValue() == "node") {
	    type = NODE;
	} else {
	    std::cerr << "Invalid type specified for object '" << getName()
		<< "'." << std::endl;
	}
	return true;
    }

    if (pname == "dynamic" && type == LIGHT) {
	// Dynamic light. Will be handled later in writeLightProperties().
	return false;
    }

    if (pname == "portal") {
	portal = property->getValue();
	type=PORTAL;
	return true;
    }

    if (pname == "portalflags") {
	const std::string& flags = property->getValue();
	portalflags = 0;
	if (flags.find("zfill") != std::string::npos)
	    portalflags |= PORTAL_ZFILL;
	if (flags.find("float") != std::string::npos)
	    portalflags |= PORTAL_FLOAT;
	if (flags.find("clip") != std::string::npos)
	    portalflags |= PORTAL_CLIP;
	if (flags.find("mirror") != std::string::npos)
	    portalflags |= PORTAL_MIRROR;
	if (flags.find("static") != std::string::npos)
	    portalflags |= PORTAL_STATIC;
	if (flags.find("autoresolve") != std::string::npos)
	    portalflags |= PORTAL_AUTORESOLVE;
	if (flags.find("clipstraddling") != std::string::npos)
	    portalflags |= PORTAL_CLIPSTRADDLING;
	if (flags.find("viscull") != std::string::npos)
	    portalflags |= PORTAL_VISCULL;
	return true;
    }

    if (pname == "portalmaxvisit") {
        std::stringstream str(property->getValue());
        str >> portalmaxvisit;
        return true;
    }
    
    if (pname == "portalwarpbefore") {
	try {
	    portalwarpbefore.scanString(property->getValue().c_str());
	} catch (dna::Error& e) {
	    std::cerr << "Error parsing portalwarpbefore property in Object '"
		<< name << "' :" << e.what() << "\n";
	}
	return true;
    }

    if (pname == "portalwarpafter") {
	try {
	    portalwarpafter.scanString(property->getValue().c_str());
	} catch (dna::Error& e) {
	    std::cerr << "Error parsing portalwarpafter property in Object '"
		<< name << "' :" << e.what() << "\n";
	}
	return true;
    }
    
    if (pname == "portalrotate") {
	try {
	    portalrotate.scanString(property->getValue().c_str());
	} catch (dna::Error& e) {
	    std::cerr << "Error parsing portalrotate property in Object '"
		<< name << "' :" << e.what() << "\n";
	}
	return true;
    }

    if (pname == "priority") {
	flags.push_back(new Flag("priority", property->getValue()));
	return true;
    }

    if (pname == "factory") {
	flags.push_back(new Flag("factory", property->getValue()));
	return true;
    }

    if (pname == "zmode") {
	zmode = ZUSE;
	if (property->getValue() == "zuse") {
	    zmode = ZUSE;
	} else if (property->getValue() == "zfill") {
	    zmode = ZFILL;
	} else if (property->getValue() == "ztest") {
	    zmode = ZTEST;
	} else if (property->getValue() == "znone") {
	    zmode = ZNONE;
	} else {
	    std::cerr << "Invalid zmode specified for object '" << name <<
			 "' using zuse instead." << std::endl;
	}
	return true;
    }

    if (pname == "mixmode") {
	meshflags.mixmode = meshflags.COPY;
	if (property->getValue() == "copy") {
	    meshflags.mixmode = meshflags.COPY;
	} else if (property->getValue() == "add") {
	    meshflags.mixmode = meshflags.ADD;
	} else if (property->getValue() == "multiply") {
	    meshflags.mixmode = meshflags.MULTIPLY;
	} else if (property->getValue() == "multiply2") {
	    meshflags.mixmode = meshflags.MULTIPLY2;
	} else if (property->getValue() == "alpha") {
	    meshflags.mixmode = meshflags.ALPHA;
	} else {
	    std::cerr << "Invalid mixmode specified for object '" << name <<
			 "' using COPY instead." << std::endl;
	}
	return true;
    }
    if (pname == "mixmode_alpha") {
	meshflags.mixmode = meshflags.ALPHA;
	meshflags.alpha = property->getValueFloat();
	return true;
    }

    PolygonFlags& polygonflags = meshflags.polygonflags;
    if (pname == "lighting") {
	parseBoolProperty(property, polygonflags.lighting);
	return true;
    }
    if (pname == "colldet") {
	parseBoolProperty(property, meshflags.colldet);
	return true;
    }
    if (pname == "viscull") {
	parseBoolProperty(property, meshflags.viscull);
	return true;
    }
    if (pname == "noclip") {
	parseBoolProperty(property, meshflags.noclip);
	return true;
    }
    if (pname == "smooth") {
        parseBoolProperty(property, meshflags.smoothing);
        return true;
    }
    if (pname == "export") {
    	parseBoolProperty(property, meshflags.exported);
	return true;
    }
    if (pname == "localshadows") {
        parseBoolProperty(property, meshflags.localshadows);
        return true;
    }
    if (pname == "hitbeam") {
	bool hitbeam = true;
	parseBoolProperty(property, hitbeam);
	if (hitbeam == false)
	    flags.push_back(new Flag("nohitbeam"));
	return true;
    }
    if (pname == "shadows") {
	bool shadows = true;
	parseBoolProperty(property, shadows);
	if (shadows == false)
	    flags.push_back(new Flag("noshadows"));
	return true;
    }
    if (pname == "camera") {
	bool camera = false;
	parseBoolProperty(property, camera);
	if (camera == true)
	    flags.push_back(new Flag("camera"));
	return true;
    }
    if (pname == "invisible") {
	bool invisible = false;
	parseBoolProperty(property, invisible);
	if (invisible == true)
	    flags.push_back(new Flag("invisible"));
	return true;
    }
    if (pname == "closed") {
	bool closed = false;
	parseBoolProperty(property, closed);
	if (closed == true)
	    flags.push_back(new Flag("closed"));
	return true;
    }
    if (pname == "goodoccluder") {
	bool goodoccluder = false;
	parseBoolProperty(property, goodoccluder);
	if (goodoccluder == true)
	    flags.push_back(new Flag("goodoccluder"));
	return true;
    }
    if (pname == "badoccluder") {
	bool badoccluder = false;
	parseBoolProperty(property, badoccluder);
	if (badoccluder == true)
	    flags.push_back(new Flag("badoccluder"));
	return true;
    }
    if (pname == "staticpos") {
	bool staticpos = false;
	parseBoolProperty(property, staticpos);
	if (staticpos == true)
	    flags.push_back(new Flag("staticpos"));
	return true;
    }
    if (pname == "staticshape") {
	bool staticshape = false;
	parseBoolProperty(property, staticshape);
	if (staticshape == true)
	    flags.push_back(new Flag("staticshape"));
	return true;
    }
    if (pname == "static") {
	bool stat = false;
	parseBoolProperty(property, stat);
	if (stat == true) {
	    flags.push_back(new Flag("staticpos"));
	    flags.push_back(new Flag("staticshape"));
	}
	return true;
    }

    TerrainMeshFlags& tflags = meshflags.terrain_flags;
    if (pname == "terrain_heightmap") {
	tflags.heightmap = property->getValue();
	return true;
    }
    if (pname == "terrain_basematerial") {
	tflags.basematerial = property->getValue();
	return true;
    }
    if (pname == "terrain_basematerial_name") {
	tflags.basematerial_name = property->getValue();
	return true;
    }
    if (pname == "terrain_materialmap") {
	tflags.materialmap = property->getValue();
	return true;
    }
    if (pname == "terrain_splatting_distance") {
	tflags.splatting_distance = property->getValueFloat();
	return true;
    }
    if (pname == "terrain_resolution") {
	tflags.resolution = property->getValueInt();
	return true;
    }
    if (pname == "terrain_block_resolution") {
	tflags.block_resolution = property->getValueInt();
	return true;
    }
    if (pname == "terrain_block_split_distance") {
	tflags.block_split_distance = property->getValueFloat();
	return true;
    }
    if (pname == "terrain_minimum_block_size") {
	tflags.minimum_block_size = property->getValueFloat();
	return true;
    }
    if (pname == "terrain_cd_resolution") {
	tflags.cd_resolution = property->getValueInt();
	return true;
    }
    if (pname == "terrain_cd_lod_cost") {
	tflags.cd_lod_cost = property->getValueFloat();
	return true;
    }
    if (pname == "terrain_lightmap_resolution") {
	tflags.lightmap_resolution = property->getValueInt();
	return true;
    }
    if (pname == "terrain_staticlighting") {
	parseBoolProperty(property, tflags.staticlighting);
	return true;
    }
    if (pname == "terrain_castshadows") {
	parseBoolProperty(property, tflags.castshadows);
	return true;
    }
    if (pname == "terrain_saveheightmap") {
	parseBoolProperty(property, tflags.saveheightmap);
	return true;
    }
    if (pname == "terrain_savematerialmap") {
	parseBoolProperty(property, tflags.savematerialmap);
	return true;
    }
    if (pname == "terrain_savebasematerial") {
	parseBoolProperty(property, tflags.savebasematerial);
	return true;
    }

    return false;
}

} // end of namespace blend2cs
