#ifndef __BLEND2CS_OBJECTS_H__
#define __BLEND2CS_OBJECTS_H__

#include <string>
#include <vector>
#include <dna++/structure.h>

#include "vectoriterator.h"
#include "vector3d.h"
#include "meshflags.h"
#include "property.h"

namespace blend2cs
{

/**
 * This class represents a CS MeshObject
 */
class Object
{
public:
    Object(const dna::Structure& structure);
    ~Object();

    void addChild(Object* object);
    
    typedef VectorIterator<Object*> ChildIterator;
    ChildIterator getChildIterator() const;
    bool hasChilds() const
    { return childs.size() > 0; }

    typedef VectorIterator<Property*> PropertyIterator;
    PropertyIterator getPropertyIterator() const;

    const Property* getProperty(const std::string& name) const;

    dna::Structure getStructure() const
    { return structure; }

    dna::Structure getData() const;

    enum Type {
	NULLMESH, LIGHT, THING, PORTAL, GENMESH, SPRITE3D, CAMERA, NODE,
	MESHREF, MESHLIB, TERRAIN, UNKNOWN
    };
    
    Type getType() const
    { return type; }
    void setType(Type newtype)
    { type = newtype; }
    
    const Vector3D& getPosition() const
    { return position; }
    
    /** get rotation vector in radians */
    const Vector3D& getRotation() const
    { return rotation; }
    
    const std::string& getName() const
    { return name; }

    const Object* getParent() const
    { return parent; }

    const std::string getPortalSector() const
    { return portal; }

    /// Take care to sync these with polygon.h
    enum {
	PORTAL_ZFILL	      = 0x01,
	PORTAL_FLOAT	      = 0x02,
	PORTAL_CLIP	      = 0x04,
	PORTAL_MIRROR	      = 0x08,
	PORTAL_STATIC	      = 0x10,
        PORTAL_AUTORESOLVE    = 0x20,
        PORTAL_CLIPSTRADDLING = 0x40,
        
        PORTAL_VISCULL        = 0x100,
    };
    const int getPortalFlags() const
    { return portalflags; }
    const int getPortalMaxVisit() const
    { return portalmaxvisit; }
    const Vector3D& getPortalRotate() const
    { return portalrotate; }
    const Vector3D& getPortalWarpBefore() const
    { return portalwarpbefore; }
    const Vector3D& getPortalWarpAfter() const
    { return portalwarpafter; }

    enum ZMode
    { ZUSE, ZNONE, ZFILL, ZTEST };
    ZMode getZMode() const
    { return zmode; }

    const MeshFlags& getMeshFlags() const
    { return meshflags; }
    
    struct Flag
    {
	Flag(const char* newname)
	    : name(newname)
	{ }
	Flag(const std::string& flagname, const std::string& flagvalue)
	    : name(flagname), value(flagvalue)
	{ }

	std::string name;
	std::string value;
    };
    typedef VectorIterator<Flag*> FlagIterator;
    FlagIterator getFlagIterator() const;

private:
    void parseObject();

    bool interpretProperty(const Property* property);
    void parseBoolProperty(const Property* property, bool& var);
    
    dna::Structure structure;

    // parsed data
    std::string name;
    Type type;

    // transform
    Vector3D position;
    Vector3D rotation;
    // note: you can't scale objects in CS

    // portal properties
    std::string portal;
    int portalflags;
    int portalmaxvisit;
    Vector3D portalrotate;
    Vector3D portalwarpbefore;
    Vector3D portalwarpafter;
    ZMode zmode;
    // other flags
    std::vector<Flag*> flags;
        
    MeshFlags meshflags;
    
    std::vector<Object*> childs;
    Object* parent;

    std::vector<Property*> properties;
};

} // end of namespace blend2cs

#endif

