#include <config.h>

#include <string.h>
#include <math.h>
#include <dna++/error.h>

#include "polygon.h"
#include "xmlwriter.h"

namespace blend2cs
{

Polygon::Polygon()
{
  clearPoly();
}
  
void Polygon::clearPoly()
{
  // Set some default inital values.
  material="no";
  flags = PolygonFlags();
  vertexcount = 0;
  memset(vertices, 0, sizeof(vertices));
  memset(uv, 0, sizeof(uv));
}

Polygon::Polygon(dna::Structure mface, dna::Structure tface,
		 const PolygonFlags& newflags, int skip)     
{
  // This polygon constructor does parsing. (Used to be Mesh::parsePolygon)
  // If polyn is 0, this polygon is either a triangle, or it is the first
  // half of a quad.  In either case, use vertices v1,v2, & v3.
  // if polyn is 1, this polygon is the second half of a quad, and we use
  // v3, v4, and v1, as per blendconv
  int i, j;

  clearPoly();  // Reset certain values.

  char str[4];
  for (i=4; i>0; i--) {
    if (i==skip) continue;      // Skip this vertex.
    sprintf(str,"v%d",i);  // Create v1, v2, v3, v4 strings
    addVertex(mface.getVariable(str).getValueInt());
  }

  flags = newflags;

  if (!tface.isGood()) {  // if Tface structure doesn't exist
    // set some boring default mapping
    for (int i=4, j=0; i>0; i--) {
      if (i==skip) continue;      // Skip this vertex.
      setUV(j,i>2 ? 0:1 ,(i==1|| i==4)? 0:1 );
      j++;
    }

    return;
  }

  dna::Variable uv = tface.getVariable("uv");
  // some settings
  short faceflags = tface.getVariable("mode").getValueShort();
  // override default colldet flag
  if(flags.colldet)
      flags.colldet = faceflags & TF_DYNAMIC;
  
  // invert v coordinat for CS
  for (i=4, j=0; i>0; i--) {
    if (i==skip) continue;      // Skip this vertex.
    setUV(j, uv.getValueFloat(2*i-2), 1.-uv.getValueFloat(2*i-1));
    j++;
  }

  fixUV();
  dna::Structure imagestruct 
    = tface.getVariable("tpage").getValueStructure();
  if (imagestruct.isGood()) {
    setMaterial( imagestruct.getVariable("id.name[24]").getValueString()+2);
    // +2 takes off the IM
  }
}

Polygon::Polygon(const Polygon& other)
    : material(other.material), flags(other.flags)
{
    vertexcount = other.vertexcount;
    memcpy(vertices, other.vertices, sizeof(vertices));
    memcpy(uv, other.uv, sizeof(uv));
}

Polygon::~Polygon()
{
}

Polygon* Polygon::createInvertedPolygon() const
{
    Polygon* polygon = new Polygon(*this);
    for(int i=vertexcount-1; i>=0; i--) {
	polygon->vertices[vertexcount-1-i]=vertices[i];
	polygon->uv[vertexcount-1-i][0]=uv[i][0];
	polygon->uv[vertexcount-1-i][1]=uv[i][1];
    }

    return polygon;
}

void Polygon::fixUV()
{
#if 0
    float minu = uv[0][0];
    for (int i=1;i<3;i++)
	if (uv[i][0] < minu)
	    minu = uv[i][0];
    // fix negative values...
    if (minu < 0) {
	float addi = -((int) minu) + 1;
	for (int i=0;i<3;i++) uv[i][0] += addi;
    }
    
    float minv = uv[0][1];
    for (int i=1;i<3;i++)
	if (uv[i][1] < minv)    
	    minv = uv[i][1];    

    // fix negative values...
    if (minv < 0) {
	float addi = -((int) minv) + 1;
	printf ("Addi: %f - %f\n", minv, addi);
	for (int i=0;i<3;i++) uv[i][1] += addi;
    }
#endif
}

void
Polygon::write(XmlWriter& xml) const
{
    for(int i = 0; i<vertexcount; ++i) {
        xml.writeTag("v");
        xml.write(getVertex(i));
    }
    
    xml.writeTag("material");
    xml.write(material);
    xml.openSection("texmap");
    for(int i=0; i<3; ++i) {
        xml.writeTag("uv");
        xml.writeAttribute("idx", i);
        xml.writeAttribute("u", getU(i));
        xml.writeAttribute("v", getV(i));
    }
    xml.closeSection("texmap");

    if(flags.lighting == false) {
        xml.writeTag("lighting");
        xml.write("no");
    }
    if(flags.colldet == false) {
        xml.writeTag("colldet");
        xml.write("no");
    }
    if(flags.shading == false) {
        xml.writeTag("shading");
        xml.write("no");
    }
}

} // end of namespace blend2cs
