#include <config.h>
#include <cmath>
#include <iostream>
#include <dna++/blendertypes.h>
#include <dna++/variable.h>
#include <valarray>

#include <png.h>
#include "vector3d.h"
#include "object.h"
#include "terrainmesh.h"
#include "xmlwriter.h"
#include "../options.h"

namespace blend2cs
{

TerrainMesh::TerrainMesh(const dna::Structure& newstructure,
        const MeshFlags& newflags)
    : structure(newstructure)
{
    usecount = 2; // always write terrains as factories
    flags = newflags;
    terrain_resolution = newflags.terrain_flags.resolution;

    int totsize = (terrain_resolution+1) * (terrain_resolution+1);
    terrainfaces = new TerrainFaces[totsize];
    heightmap = new float[totsize];
    materialmap = new int[totsize];
    // @@@ Base material should be able to have independent resolution!
    basematerial = new unsigned char[totsize*3];

    parseMesh();
    generateHeightmap();
}

TerrainMesh::~TerrainMesh ()
{
    delete[] terrainfaces;
    delete[] heightmap;
    delete[] materialmap;
    delete[] basematerial;
}

/* Helper function */
void unpack_vertex_color (dna::Variable& colvar, int n, png_color& color)
{
    /* TODO - there are probably endianess problems here */
    if (n >= 0 && n < 4)
    {
	uint32 bits = (uint32)colvar.getValueInt(n);
	color.red = (bits >> 24) & 0xFF;
	color.green = (bits >> 16) & 0xFF;
	color.blue = (bits >> 8) & 0xFF;
    }
}

void TerrainMesh::makeTri(
	const std::vector<Vector3D>& vertices,
	dna::Structure mface, dna::Structure tface,
	int skip,
	Triangle3D& tri, std::string& material,
	int& red, int& green, int& blue)
{
    int i;
    int tri_idx[4];
    int idx = 0;
    char str[4];
    for (i=4; i>0; i--) {
        if (i==skip) continue;      // Skip this vertex.
	sprintf(str,"v%d",i);  // Create v1, v2, v3, v4 strings
        int vt = mface.getVariable(str).getValueInt();
        tri_idx[idx++] = vt;
    }

    material = "no";
    if (tface.isGood()) {
    	dna::Structure imagestruct 
            = tface.getVariable("tpage").getValueStructure();
	if (imagestruct.isGood()) {
	    material = imagestruct.getVariable("id.name[24]").getValueString()+2;
    	    // +2 takes off the IM
        }
    }
    tri.set(
    	vertices[tri_idx[0]],
    	vertices[tri_idx[1]],
    	vertices[tri_idx[2]]);

    png_color color;
    dna::Variable colvar = tface.getVariable("col");
    /* We know the color at each corner of the patch, so compute the 
     * average to get the color in the centre. */
    red = 0;
    green = 0;
    blue = 0;
    for (int i = 0; i < 3; i++) {
	unpack_vertex_color(colvar, i, color);
	red += color.red;
	green += color.green;
	blue += color.blue;
    }
    red /= 3;
    green /= 3;
    blue /= 3;
}

int TerrainMesh::findMaterialIdx(const std::string& material)
{
    VectorIterator<std::string> it = VectorIterator<std::string> (material_palette);
    int idx = 0;
    while (it.hasNext()) {
    	std::string mat = it.next();
	if (mat == material) return idx;
	idx++;
    }
    material_palette.push_back (material);
    return idx;
}

void TerrainMesh::distributeTri(const Triangle3D& tri,
	const std::string& material, float dx, float dy,
	int red, int green, int blue)
{
    const Vector2D& bounds_min = scanbounds.getMin();
    int material_idx = findMaterialIdx(material);

    Box2D box = tri.getBoundsXZ();
    int box_x1 = int ((box.getMin().getX()-bounds_min.getX()) / dx);
    int box_y1 = int ((box.getMin().getY()-bounds_min.getY()) / dy);
    int box_x2 = int ((box.getMax().getX()-bounds_min.getX()) / dx) + 1;
    int box_y2 = int ((box.getMax().getY()-bounds_min.getY()) / dy) + 1;
    if (box_x1 < 0) box_x1 = 0;
    if (box_y1 < 0) box_y1 = 0;
    if (box_x2 > terrain_resolution) box_x2 = terrain_resolution;
    if (box_y2 > terrain_resolution) box_y2 = terrain_resolution;

    int x, y;
    for (y = box_y1 ; y <= box_y2 ; y++) {
    	int yidx = y * (terrain_resolution+1);
        for (x = box_x1 ; x <= box_x2 ; x++) {
	    Vector2D point (x*dx + bounds_min.getX(), y*dy + bounds_min.getY());
	    if (tri.in(point)) {
	    	TerrainFace tf;
		tf.tri = tri;
		tf.material_idx = material_idx;
		tf.red = red;
		tf.green = green;
		tf.blue = blue;
	    	terrainfaces[yidx + x].push_back(tf);
	    }
	}
    }
}

void TerrainMesh::parseMesh()
{
    name = structure.getVariable("id.name").getValueString() +2; // +2 skips ME

    // parse vertices
    int vertexcount = structure.getVariable("totvert").getValueInt();
    dna::Structure firstvertex 
	= structure.getVariable("mvert").getValueStructure();

    minheight = 1000000000.0f;
    maxheight = -1000000000.0f;
    bounds.set (1000000000.0f, 1000000000.0f,
		-1000000000.0f, -1000000000.0f);
    bool needscale = flags.scale != Vector3D(0,0,0);
    dna::Structure::Iterator i = firstvertex.getIterator("MVert");
    std::vector<Vector3D> vertices;
    // First we scan all vertices to calculate the bounding box of the
    // terrain.
    while(i.hasNext() && vertexcount-- > 0) {
	dna::Structure vertex = i.next();

	dna::Variable coords = vertex.getVariable("co");
	// remember to swap y and z and to revert y here!
	Vector3D v (coords.getValueFloat(0),
		    coords.getValueFloat(2),
		    coords.getValueFloat(1));
	if (needscale)
	    v.scale(flags.scale);
	vertices.push_back(v);
	bounds.addBoundingVertex(v.getX(), v.getZ());
	if (v.getY() < minheight) minheight = v.getY();
	if (v.getY() > maxheight) maxheight = v.getY();
    }

    // Slightly scale down the boundaries horizontally so that we
    // get no problems at the borders.
    float xextend = (bounds.getMax().getX()-bounds.getMin().getX())/2;
    float yextend = (bounds.getMax().getY()-bounds.getMin().getY())/2;
    float xcenter = (bounds.getMax().getX()+bounds.getMin().getX())/2;
    float ycenter = (bounds.getMax().getY()+bounds.getMin().getY())/2;
    scanbounds.setMinX(xcenter-xextend*0.999);
    scanbounds.setMinY(ycenter-yextend*0.999);
    scanbounds.setMaxX(xcenter+xextend*0.999);
    scanbounds.setMaxY(ycenter+yextend*0.999);
    printf ("scanbounds=%g,%g    %g,%g\n", scanbounds.getMin().getX(),scanbounds.getMin().getY(),scanbounds.getMax().getX(), scanbounds.getMax().getY()); fflush (stdout);
    printf ("bounds=%g,%g    %g,%g\n", bounds.getMin().getX(),bounds.getMin().getY(),bounds.getMax().getX(), bounds.getMax().getY()); fflush (stdout);

    // Now we scan all faces and classify them in the right point in
    // our 2-dimensional terrainfaces array.
    const Vector2D& bounds_min = scanbounds.getMin();
    const Vector2D& bounds_max = scanbounds.getMax();
    float dx = (bounds_max.getX()-bounds_min.getX()) / float (terrain_resolution);
    float dy = (bounds_max.getY()-bounds_min.getY()) / float (terrain_resolution);
    int facecount = structure.getVariable("totface").getValueInt();

    dna::Structure firstface 
	= structure.getVariable("mface").getValueStructure();
    dna::Structure firstuvface 
	= structure.getVariable("tface").getValueStructure();

    dna::Structure::Iterator fi = firstface.getIterator("MFace");
    dna::Structure::Iterator u = firstuvface.getIterator("TFace");

    while(fi.hasNext() && facecount-- > 0) {
	dna::Structure mface = fi.next();
	dna::Structure tface = dna::Structure::INVALID_PTR;
	
    	if (u.hasNext()) {
	    tface = u.next(); 
	}

	// Is this a quad or a triangle?  2 is quad, 1 is triangle
	Triangle3D tri;
	std::string material;
	int red, green, blue;
	if (mface.getVariable("v4").getValueInt() > 0 ) {  // it is a quad
	    makeTri(vertices,mface,tface, 4, tri, material, red, green, blue);
	    distributeTri(tri, material, dx, dy, red, green, blue);
	    makeTri(vertices,mface,tface, 2, tri, material, red, green, blue);
	    distributeTri(tri, material, dx, dy, red, green, blue);
	} else {
	    // Triangle
	    makeTri(vertices,mface,tface, 4, tri, material, red, green, blue);
	    distributeTri(tri, material, dx, dy, red, green, blue);
        }
    }
}

void TerrainMesh::generateHeightmap()
{
    float start_y = maxheight*2;
    float end_y = minheight*2;
    Vector2D point;

    float dx = (scanbounds.getMax().getX()-scanbounds.getMin().getX()) / float (terrain_resolution);
    float dz = (scanbounds.getMax().getY()-scanbounds.getMin().getY()) / float (terrain_resolution);

    int x, z;
    int idx = 0;
    for (z = 0 ; z <= terrain_resolution ; z++) {
	point.setY (z * dz + scanbounds.getMin().getY());
        for (x = 0 ; x <= terrain_resolution ; x++) {
	    TerrainFaces& tf = terrainfaces[idx];
	    point.setX (x * dx + scanbounds.getMin().getX());
	    float maxh = minheight*2;
	    int matidx = -1;
	    float isect_y;
	    VectorIterator<TerrainFace> it = VectorIterator<TerrainFace> (tf);
	    int red = 0;
	    int green = 0;
	    int blue = 0;
	    int color_count = 0;
	    while (it.hasNext()) {
	    	TerrainFace face = it.next();
		if (face.tri.intersectVert(point, start_y, end_y, isect_y)) {
		    color_count++;
		    red += face.red;
		    green += face.green;
		    blue += face.blue;
		    if (isect_y > maxh) {
		        maxh = isect_y;
			matidx = face.material_idx;
		    }
		}
	    }
	    if (matidx == -1) {
		std::cerr << "Error! Missing geometry for heightmap at " << x << "," << z << "!" << std::endl;
		matidx = 0;
		maxh = (maxheight-minheight) / 2.0f;
	    }
	    heightmap[idx] = maxh;
	    materialmap[idx] = matidx;
	    if (color_count > 0) {
	        red /= color_count;
	        green /= color_count;
	        blue /= color_count;
	    }
	    basematerial[idx*3+0] = red;
	    basematerial[idx*3+1] = green;
	    basematerial[idx*3+2] = blue;

	    idx++;
	}
    }
}

void TerrainMesh::write(XmlWriter& xml) const
{
    xml.openSection("params");
    xml.writeTag("plugin");
    xml.write("crystalspace.mesh.object.terrain.bruteblock");
    xml.writeTag("terraformer");
    xml.write(getName());

    xml.openSection("sampleregion");
    xml.writeTag("min");
    xml.writeAttribute("x", bounds.getMin().getX());
    xml.writeAttribute("y", bounds.getMin().getY());
    xml.writeTag("max");
    xml.writeAttribute("x", bounds.getMax().getX());
    xml.writeAttribute("y", bounds.getMax().getY());
    xml.closeSection("sampleregion");

    xml.closeSection("params");
}

struct datastore{
  unsigned char *data;
  size_t pos;
  size_t length;

  datastore () { data = 0; pos = 0; length = 0; }
  ~datastore () { free (data); }
};

static void png_write (png_structp png, png_bytep data, png_size_t length)
{
  datastore *ds = (datastore *)png->io_ptr;
  if (ds->pos + (long)length > ds->length)
  {
    ds->data = (unsigned char*)realloc (ds->data, ds->pos + (long)length);
    if (!ds->data)
      png_error (png, "memory allocation error");
    else
      ds->length = ds->pos + length;
  }
  memcpy (ds->data + ds->pos, data, length);
  ds->pos += length;
}

void png_flush (png_structp)
{
}

bool TerrainMesh::saveHeightMap(const std::string& path) const
{
    FILE *file;
    png_structp pngptr;
    png_infop info;

    /* Create the PNG structure */
    pngptr = png_create_write_struct (PNG_LIBPNG_VER_STRING,
				      NULL, NULL, NULL);
    if (pngptr == NULL)
	return false;

    info = png_create_info_struct (pngptr);
    if (info == NULL)
    {
	png_destroy_write_struct(&pngptr, &info);
	return false;
    }

    const char* str = (path+flags.terrain_flags.heightmap).c_str();
    file = fopen(str, "wb");
    if (file == NULL)
    {
	std::cerr << "cannot open " << str << " for writing" << std::endl;
	png_destroy_write_struct(&pngptr, &info);
	return false;
    }
    datastore outfun;
    png_set_write_fn(pngptr, (png_voidp)&outfun, png_write, png_flush);

    /* Initialize the header */
    int rows = terrain_resolution+1;
    int cols = terrain_resolution+1;
    png_set_IHDR (pngptr, info, cols, rows, 8, // 8-bpp
		  PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE,
		  PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

    /* Now stuff the bytes and write the PNG file */
    png_bytepp data = new png_bytep[rows];
    int idx = 0;
    for (int r = 0; r < rows; r++)
    {
        int dr = rows-r-1;
	data[dr] = new png_byte[cols];
	for (int c = 0; c < cols; c++)
	{
	    int h = (png_byte)(256.0 * (heightmap[idx]-minheight) / (maxheight-minheight));
	    if (h > 255) h = 255;
	    else if (h < 0) h = 0;
	    data[dr][c] = h;
	    idx++;
	}
    }
    png_write_info(pngptr, info);
    png_write_image (pngptr, data);
    png_write_end (pngptr, info);
    /* Free memory */
    for (int r = 0; r < rows; r++) delete data[r];
    delete[] data;
    png_destroy_write_struct(&pngptr, &info);

    fwrite(outfun.data, 1, outfun.length, file);
    fclose (file);

    return true;
}

bool TerrainMesh::saveBaseMaterial(const std::string& path) const
{
    FILE *file;
    png_structp pngptr;
    png_infop info;

    /* Create the PNG structure */
    pngptr = png_create_write_struct (PNG_LIBPNG_VER_STRING,
				      NULL, NULL, NULL);
    if (pngptr == NULL)
	return false;

    info = png_create_info_struct (pngptr);
    if (info == NULL)
    {
	png_destroy_write_struct(&pngptr, &info);
	return false;
    }

    const char* str = (path+flags.terrain_flags.basematerial).c_str();
    file = fopen(str, "wb");
    if (file == NULL)
    {
	std::cerr << "cannot open " << str << " for writing" << std::endl;
	png_destroy_write_struct(&pngptr, &info);
	return false;
    }
    datastore outfun;
    png_set_write_fn(pngptr, (png_voidp)&outfun, png_write, png_flush);

    /* Initialize the header */
    int rows = terrain_resolution+1;
    int cols = terrain_resolution+1;
    png_set_IHDR (pngptr, info, cols, rows, 8,
		  PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
		  PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

    /* Now stuff the bytes and write the PNG file */
    png_bytep data[rows];
    for (int r = 0; r < rows; r++)
    {
        int dr = rows-r-1;
	data[dr] = (png_bytep)(&basematerial[r*cols*3]);
    }
    png_write_info(pngptr, info);
    png_write_image (pngptr, data);
    png_write_end (pngptr, info);

    png_destroy_write_struct(&pngptr, &info);

    fwrite(outfun.data, 1, outfun.length, file);
    fclose (file);

    return true;
}

bool TerrainMesh::saveMaterialPalette(const std::string& path) const
{
    FILE *file;
    png_structp pngptr;
    png_infop info;

    /* Create the PNG structure */
    pngptr = png_create_write_struct (PNG_LIBPNG_VER_STRING,
				      NULL, NULL, NULL);
    if (pngptr == NULL)
	return false;

    info = png_create_info_struct (pngptr);
    if (info == NULL)
    {
	png_destroy_write_struct(&pngptr, &info);
	return false;
    }

    const char* str = (path+flags.terrain_flags.materialmap).c_str();
    file = fopen(str, "wb");
    if (file == NULL)
    {
	std::cerr << "cannot open " << str << " for writing" << std::endl;
	png_destroy_write_struct(&pngptr, &info);
	return false;
    }
    datastore outfun;
    png_set_write_fn(pngptr, (png_voidp)&outfun, png_write, png_flush);

    /* Initialize the header */
    int rows = terrain_resolution+1;
    int cols = terrain_resolution+1;
    png_set_IHDR (pngptr, info, cols, rows, 8, 
		  PNG_COLOR_TYPE_PALETTE, PNG_INTERLACE_NONE,
		  PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

    /* Define the palette. Assign a unique color to each material in the map */
    int len = material_palette.size();
    png_color *palette = new png_color[len];
    for (int i = 0; i < len; i++)
    {
	png_color color;
	/* This will produce up to 255 "random" and distinct colors (nice 
	 * to have if you need to debug things). */
	color.red = (31*i) % 255;
	color.green = (63*i) % 255;
	color.blue = (127*i) % 255;
	palette[i] = color;
    }
    png_set_PLTE (pngptr, info, palette, len);

    /* Now stuff the bytes and write the PNG file */
    png_bytepp data = new png_bytep[rows];
    int idx = 0;
    for (int r = 0; r < rows; r++)
    {
        int dr = rows-r-1;
	data[dr] = new png_byte[cols];
	for (int c = 0; c < cols; c++)
	{
	    data[dr][c] = (png_byte)materialmap[idx];
	    idx++;
	}
    }
    png_write_info(pngptr, info);
    png_write_image (pngptr, data);
    png_write_end (pngptr, info);
    /* Free memory */
    for (int r = 0; r < rows; r++) delete data[r];
    delete[] data;
    png_destroy_write_struct(&pngptr, &info);

    fwrite(outfun.data, 1, outfun.length, file);
    fclose (file);

    return true;
}


} // end of namespace blend2cs
