#include <config.h>
#include <cmath>
#include <iostream>
#include <dna++/blendertypes.h>
#include <dna++/variable.h>
#include <valarray>

#include "vector3d.h"
#include "object.h"
#include "thingmesh.h"
#include "xmlwriter.h"
#include "../options.h"

namespace blend2cs
{

ThingMesh::ThingMesh(const dna::Structure& newstructure,
        const MeshFlags& newflags, bool dummy, bool eliminateEmptyTriangles,
	bool dontWarnTexture)
    : structure(newstructure), supportfactory(true)
{
    flags = newflags;

    if(dummy) {
        // Use this constructor to create a mesh that is a dummy mesh...
        name= structure.getVariable("name").getValueString() +2;
        usecount = 2;
    } else {
        parseMesh(eliminateEmptyTriangles, dontWarnTexture);
    }
}

ThingMesh::~ThingMesh ()
{
    PolygonIterator p = getPolygonIterator();
    while(p.hasNext())
	delete p.next();
}

ThingMesh::VertexIterator ThingMesh::getVertexIterator() const
{
    return VertexIterator(vertices);
}

ThingMesh::PolygonIterator ThingMesh::getPolygonIterator() const
{
    return PolygonIterator(polygons);
}

void ThingMesh::parseMesh(bool eliminateEmptyTriangles, bool dontWarnTexture)
{
    name = structure.getVariable("id.name").getValueString() +2; // +2 skips ME
    
    parseVertices();
    parseFaces(eliminateEmptyTriangles, dontWarnTexture);
}

void ThingMesh::parseVertices()
{
    //parse vertices
    int vertexcount = structure.getVariable("totvert").getValueInt();
    dna::Structure firstvertex 
	= structure.getVariable("mvert").getValueStructure();

    dna::Structure::Iterator i = firstvertex.getIterator("MVert");
    while(i.hasNext() && vertexcount-- > 0) {
	dna::Structure vertex = i.next();

	dna::Variable coords = vertex.getVariable("co");
	// remember to swap y and z and to revert y here!
	vertices.push_back(Vector3D(
		    coords.getValueFloat(0),
		    coords.getValueFloat(2),
		    coords.getValueFloat(1)));

	// XXX: the variable "no" seems to contain normal data. But i have no
	// idea what format they are in and how to use them in CS.
    }

    // do we need to apply scaling?
    if(flags.scale != Vector3D(0,0,0)) {
        for(Vertices::iterator i = vertices.begin(); i != vertices.end(); ++i) {
            Vector3D& v = *i;
            v.scale(flags.scale);
        }
    }
}

void ThingMesh::parseFaces(bool eliminateEmptyTriangles, bool dontWarnTexture)
{
    int facecount = structure.getVariable("totface").getValueInt();

    dna::Structure firstface 
	= structure.getVariable("mface").getValueStructure();
    dna::Structure firstuvface 
	= structure.getVariable("tface").getValueStructure();
    bool hasuv = firstuvface.isGood();

    dna::Structure::Iterator i = firstface.getIterator("MFace");
    dna::Structure::Iterator u;
    if (hasuv)
	u = firstuvface.getIterator("TFace");
    else {
    	if (!dontWarnTexture)
	   std::cerr << "warning: mesh " << name << " has no UV"
	       << " coordinates assigned, using dummy texture." << std::endl;
    }

    while(i.hasNext() && facecount-- > 0) {
	dna::Structure mface = i.next();
	dna::Structure tface = dna::Structure::INVALID_PTR;
	
    	if (hasuv && u.hasNext()) {
	    tface = u.next(); 
	} else {
	    hasuv = false;
	}

	// Is this a quad or a triangle?  2 is quad, 1 is triangle
	if (mface.getVariable("v4").getValueInt() > 0 ) {  // it is a quad
	  if (!badUV(tface,mface)) {
	    makePolys(mface,tface, 0, eliminateEmptyTriangles); // UV mapping is good, make 1 poly
	  } else {  // quad that we must break in 2
	    makePolys(mface,tface, 4, eliminateEmptyTriangles);
	    makePolys(mface,tface, 2, eliminateEmptyTriangles);
	  }
	} else {
	  makePolys(mface,tface, 4, eliminateEmptyTriangles);  // Triangle
        }
    }
}

static float SquaredNorm (const Vector3D& v)
{
    return v.getX() * v.getX() + v.getY() * v.getY() + v.getZ() * v.getZ();
}

static float Area3 (const Vector3D &a, const Vector3D &b,
                             const Vector3D &c)
{
    Vector3D v1 = b; v1 -= a;
    Vector3D v2 = c; v2 -= a;
    Vector3D cross = v1 % v2;
    float sqnorm = SquaredNorm (cross);
    return sqrt(sqnorm);
}

bool
ThingMesh::emptyPolygon(const Polygon* polygon)
{
    // First check if the polygon is not completely flat (i.e. no
    // area).
    float totalarea = 0.0f;
    int p0 = polygon->getVertex(0);
    for (int i = 1; i<=polygon->getVertexCount() - 2; ++i) {
        int pi = polygon->getVertex(i);
        int pii = polygon->getVertex(i+1);
        totalarea += fabs (Area3 (getVertex(p0),
            getVertex(pi), getVertex(pii)));
    }

    return totalarea <= .0001;
}

void ThingMesh::makePolys(dna::Structure mface, dna::Structure tface, int skip,
        bool eliminateEmptyTriangles)
{
  // If poly is a quad, we make 2 polygons, each a triangle.
  // We do this because CS only handle 3 uv coordinates perface.
  
  Polygon* polygon = new Polygon(mface, tface, flags.polygonflags, skip);
  if(eliminateEmptyTriangles && emptyPolygon(polygon)) {
      delete polygon;
      std::cerr << "Skipping polygon in mesh '" << name
            << "' because it is empty!\n";
      return;
  }
  
  polygons.push_back(polygon);
  
  if (tface.isGood()) {
    short faceflags = tface.getVariable("mode").getValueShort();
    if (faceflags & TF_TWOSIDE) {
      // It's a double sided poly, add another one for CS
      Polygon* invertedpolygon = polygon->createInvertedPolygon();
      polygons.push_back(invertedpolygon);
    }
  } 
}

int findAMax(V3DArrayType ar, unsigned int skip)
{ // This routine finds which index has the maximum value of a valarray
  unsigned int i, wmax=500;
  double max=-3e38;
  for (i=0; i< ar.size(); i++) {
    if (i==skip) continue;
    if (ar[i]>max) {
      wmax=i;
      max=ar[i];
    }
  }
  return wmax;
}

bool ThingMesh::badUV(dna::Structure tface,dna::Structure mface)
{ // This routine tells us if we have to break a quad into triangles
  // for more accurate UV mapping.  Using a linear extrapolation it 
  // estimates the UV coordinates of the 3rd vertex  based on the
  // coordinates of the other 3 vertices.  If these differ significantly
  // from the blender UV coordinates, we return true.

  int i;

  if (tface.isGood()) {
    
    dna::Variable uv = tface.getVariable("uv");  // Blender UV values.
    std::valarray<double> u(4), v(4);
    u[0]=uv.getValueFloat(0);  v[0]=uv.getValueFloat(1);
    u[1]=uv.getValueFloat(2);  v[1]=uv.getValueFloat(3);
    u[2]=uv.getValueFloat(4);  v[2]=uv.getValueFloat(5);
    u[3]=uv.getValueFloat(6);  v[3]=uv.getValueFloat(7);

    // find vertices, find x/y/z of vertices
    // vxyz[vertex #][dimension (x=0,y=1,z=2)]
    V3DArrayType vxyz[4];
    
    for (i=0; i< 4; i++ ) {  // Get Coordinates of all 4 vertices.
      char str[4];
      sprintf(str,"v%d",i+1);  // Create v1, v2, v3, v4 strings.
      vxyz[i].resize(3);
      vxyz[i]=vertices[mface.getVariable(str).getValueInt()].getValArray();
      //      for (int j=0; j<3; j++)  std::cerr << (*vxyz[i])[j] << " " ;
      //      std::cerr << std::endl;
    }

    // Numerically it would be most accurate would be to calculate the
    // linear extrapolation of uv coordinates in the plane of the
    // polygon.  Rather than do that, we just calculate it using 2 of
    // the xyz axes. Find along which axis the polygon/face might
    // lie. Don't use that one for our calculations when
    // extrapolating.
    V3DArrayType del41=((vxyz[3])-(vxyz[0]));
    V3DArrayType del21=((vxyz[1])-(vxyz[0]));
    V3DArrayType del31=((vxyz[2])-(vxyz[0]));

    int ind1=findAMax(std::abs(del21),4);  
    int ind2=findAMax(std::abs(del31),ind1);
    if (ind1<0 || ind1>2||ind2<0 || ind1>2 || ind1==ind2)     // Stupid check
      throw dna::Error ("Got an abnormal minimum index.");
    
    double denom=del41[ind2]*(double) del21[ind1]-del41[ind1]*del21[ind2];
    if (!denom)  return true;
    
    double t3=(del31[ind2]*(double) del21[ind1]-del21[ind2]*del31[ind1])/denom;
    if (fabs(t3)> 1.e5) return true;
    
    if (!del21[ind1])   return true;

    double s3=(del31[ind1]-t3*del41[ind1])/del21[ind1];
    if (fabs(s3)> 1.e5) return true;

    // Estimate of u3, v3
    float eu3= s3*(u[1]-u[0])+t3*(u[3]-u[0])+u[0];
    float ev3= s3*(v[1]-v[0])+t3*(v[3]-v[0])+v[0];
      
    // absolute Error between our linear estimate and the value from blender.
    float edu=fabs(u[2]-eu3);
    float edv=fabs(v[2]-ev3);
    
    // get rough scale of our uv values.
    double du=u.max()-u.min();
    double dv= v.max()-v.min();
    
    // If our error is more than 2% (?), split the quad.
    double urat=edu/du;
    double vrat=edv/dv;
    // fprintf(stdout, "%f %f    %f %f    %f %f",eu3,ev3,u[2],v[2],urat,vrat);
    if (urat > 0.02 || vrat>0.02 || !finite(urat) || !finite(vrat)) {
      return true;  
    }
  }    
  return false;  // No, we don't need to split this polygon.
}

void
ThingMesh::write(XmlWriter& xml) const
{
    bool colldet_disable = false;

    xml.openSection("params");

    // write vertices
    VertexIterator i = getVertexIterator();
    while(i.hasNext()) {
        const Vector3D& vertex = i.next();
        vertex.write(xml, "v");
    }

    // write polygons
    static int polnum = 0; // ad-hoc hack
    int n = 0;
    PolygonIterator p = getPolygonIterator();
    while(p.hasNext()) {
        const Polygon* polygon = p.next();

        xml.openSection("p");
        if(config.polygonnames) {
            char polyname[24];
            snprintf(polyname, 24, "%d-%d", n++, polnum++);
            xml.writeAttribute("name", polyname);
        }

        polygon->write(xml);
        xml.closeSection("p");
	if (polygon->getFlags().colldet == false)
	    colldet_disable = true;
    }
    
    if(flags.smoothing == true)
        xml.writeTag("smooth");
    writeMixMode(xml);

    xml.closeSection("params");

    if (colldet_disable) {
	   std::cerr << "warning: mesh " << name << " has polygons with colldet disabled!" << std::endl;
    }
}

} // end of namespace blend2cs
