#include <config.h>

#include <stdio.h>

#include "triangle3d.h"
#include "dna++/error.h"

namespace blend2cs
{

bool segmentPlane(
	const Vector3D& u,
	const Vector3D& v,
	const Vector3D& norm, float dd,
	Vector3D& isect)
{
    Vector3D uv = u - v;
    float denom = norm * uv;
    if (denom == 0) {
    	// they are parallel
    	isect = v;
    	return false;
    }
    float dist = (norm * u + dd) / denom;
    if (dist < -.000001 || dist > 1 + .000001) {
    	isect.set (0, 0, 0);
    	return false;
    }

    isect = u - dist * uv;
    return true;
}

bool segmentPlaneVert(
	const Vector2D& point,
	float u_y,
	float v_y,
	const Vector3D& norm, float dd,
	float& isect_y)
{
    float uv_y = u_y - v_y;
    float denom = norm.getY() * uv_y;
    if (denom == 0) {
    	// they are parallel
    	isect_y = v_y;
    	return false;
    }
    float dist = (norm.getX()*point.getX() + norm.getY()*u_y + norm.getZ()*point.getY() + dd) / denom;

    isect_y = u_y - dist * uv_y;
    return true;
}

int whichSide3D (const Vector3D& p, const Vector3D& v1, const Vector3D& v2)
{
    float s = p.getX()*(v1.getY()*v2.getZ()-v1.getZ()*v2.getY())
    	    + p.getY()*(v1.getZ()*v2.getX()-v1.getX()*v2.getZ())
	    + p.getZ()*(v1.getX()*v2.getY()-v1.getY()*v2.getX());
    if (s < -0.01) return 1;
    else if (s > 0.01) return -1;
    else return 0;
}

int whichSide2D (const Vector2D& v, const Vector2D& s1, const Vector2D& s2)
{
    float k  = (s1.getY() - v.getY())*(s2.getX() - s1.getX());
    float k1 = (s1.getX() - v.getX())*(s2.getY() - s1.getY());
    if (k < k1) return -1;
    else if (k > k1) return 1;
    else return 0;
}

int whichSide3D2D (const Vector2D& v, const Vector3D& s1, const Vector3D& s2)
{
    float k  = (s1.getZ() - v.getY())*(s2.getX() - s1.getX());
    float k1 = (s1.getX() - v.getX())*(s2.getZ() - s1.getZ());
    k -= k1;
    if (k < -0.01) return -1;
    else if (k > 0.01) return 1;
    else return 0;
}

bool Triangle3D::intersect(const Vector3D& v1, const Vector3D& v2,
	Vector3D& isect)
{
    Vector3D norm = (p1-p2) % (p1-p3);
    float dd = -(norm * p1);

    if (!segmentPlane (v1, v2, norm, dd, isect))
    	return false;

    int test1, test2, test3;
    if (dd > .00001) {
    	test1 = whichSide3D(isect, p3, p1);
    	test2 = whichSide3D(isect, p1, p2);
    	test3 = whichSide3D(isect, p2, p3);
    } else {
        // Plane is too close to origin.
	Vector3D nsect = isect + norm;
	Vector3D np1 = p1 + norm;
	Vector3D np2 = p2 + norm;
	Vector3D np3 = p3 + norm;
    	test1 = whichSide3D(isect, np3, np1);
    	test2 = whichSide3D(isect, np1, np2);
    	test3 = whichSide3D(isect, np2, np3);
    }

    // Check if the point is on the same side of each plane.
    // This works for both backface and frontface triangles.
    // Return success if all WhichSide3D tests are either all -1 or all 1.
    // Note: if we want to only check for one side of the triangle
    // and we did a plane shift, we would need to check if the same
    // face of the shifted plane is facing the origin (0,0,0).
    // Shifting the plane often times causes us to look at the other
    // side of the plane we are testing.
    if (test1 >= 0 && test2 >= 0 && test3 >= 0)
      return true;
    if (test1 <= 0 && test2 <= 0 && test3 <= 0)
      return true;
    return false;
}

bool Triangle3D::intersectVert(const Vector2D& point, float ystart, float ystop, float& yisect)
{
    Vector3D norm = (p1-p2) % (p1-p3);
    float dd = -(norm * p1);

    if (!segmentPlaneVert (point, ystart, ystop, norm, dd, yisect))
    	return false;

    int test1 = whichSide3D2D(point, p3, p1);
    int test2 = whichSide3D2D(point, p1, p2);
    int test3 = whichSide3D2D(point, p2, p3);

    // Check if the point is on the same side of each plane.
    // This works for both backface and frontface triangles.
    // Return success if all WhichSide3D tests are either all -1 or all 1.
    // Note: if we want to only check for one side of the triangle
    // and we did a plane shift, we would need to check if the same
    // face of the shifted plane is facing the origin (0,0,0).
    // Shifting the plane often times causes us to look at the other
    // side of the plane we are testing.
    if (test1 >= 0 && test2 >= 0 && test3 >= 0)
      return true;
    if (test1 <= 0 && test2 <= 0 && test3 <= 0)
      return true;
    return false;
}

Box2D Triangle3D::getBoundsXZ() const
{
    Box2D box;
    box.addBoundingVertex(p1.getX(), p1.getZ());
    box.addBoundingVertex(p2.getX(), p2.getZ());
    box.addBoundingVertex(p3.getX(), p3.getZ());
    return box;
}

bool Triangle3D::in(const Vector2D& v) const
{
    Vector2D pp1 (p1.getX(), p1.getZ());
    Vector2D pp2 (p2.getX(), p2.getZ());
    Vector2D pp3 (p3.getX(), p3.getZ());
    int s1 = whichSide2D(v, pp3, pp1);
    int s2 = whichSide2D(v, pp1, pp2);
    int s3 = whichSide2D(v, pp2, pp3);
    if (s1 <= 0 && s2 <= 0 && s3 <= 0) return true;
    if (s1 >= 0 && s2 >= 0 && s3 >= 0) return true;
    return false;
}

} // end of namespace blend2cs
