#include <config.h>

#include <iostream>
#include <map>
#include "trianglemesh.h"
#include "thingmesh.h"
#include "polygon.h"
#include "xmlwriter.h"

namespace blend2cs
{

TriangleMesh::TriangleMesh(const ThingMesh& mesh)
    : structure(mesh.getStructure())
{
    usecount = 2; // always write genmeshes as factories
    flags = mesh.getFlags();
    convertMesh(mesh);
}

TriangleMesh::~TriangleMesh()
{
}

TriangleMesh::TriangleIterator TriangleMesh::getTriangleIterator() const
{
    return TriangleIterator(triangles);
}

TriangleMesh::VertexIterator TriangleMesh::getVertexIterator() const
{
    return VertexIterator(vertices);
}

bool
TriangleMesh::supportFactory() const
{
    return true;
}

/* Algorithm to convert polygons to CS style triangles. The tricky part here is
 * that the CS vertices have UV coordinates assigned instead of the triangles.
 * So we have to eventually duplicate some vertices with different UV coords
 * here...
 */
class MeshConverter
{
public:
    MeshConverter(TriangleMesh& newtrimesh, const ThingMesh& newmesh)
	: trimesh(newtrimesh), mesh(newmesh), newidx(0)
    {
	convertMesh();
    }

    typedef std::multimap<int, int> VertexMap;

private:
    TriangleMesh& trimesh;
    const ThingMesh& mesh;
    VertexMap map;
    int newidx;
    
    /* check if we already have a vertice with certain u,v coords which
     * corresponds to an old vertex.
     */
    int createVertexIndex(int oldv, float u, float v) {
	VertexMap::iterator i = map.find(oldv);

	for( ; i != map.end() && i->first == oldv; i++) {
	    const TriangleMesh::Vertex& vertex = trimesh.vertices[i->second];
	    if (vertex.u == u && vertex.v == v)               	
		return i->second;
	}
	
	const Vector3D& vector = mesh.getVertex(oldv);
    	trimesh.vertices.push_back(TriangleMesh::Vertex(vector, u, v));
	map.insert (std::make_pair(oldv, newidx));
	return newidx++;
    }

    void convertTriangle(const Polygon* polygon, int p1, int p2, int p3)
    {
	int v1 = createVertexIndex(polygon->getVertex(p1),
				   polygon->getU(p1), polygon->getV(p1));
	int v2 = createVertexIndex(polygon->getVertex(p2),
                        	   polygon->getU(p2), polygon->getV(p2));
	int v3 = createVertexIndex(polygon->getVertex(p3),
	                 	   polygon->getU(p3), polygon->getV(p3));
	trimesh.triangles.push_back(TriangleMesh::Triangle(v1, v2, v3));
    }

    void convertMesh()
    {
	ThingMesh::PolygonIterator i = mesh.getPolygonIterator();
	while(i.hasNext()) {
	    const Polygon* polygon = i.next();
	    convertTriangle(polygon, 0, 1, 2);
	    if (polygon->getVertexCount() > 3)
		convertTriangle(polygon, 2, 3, 0);

	    if (polygon->getMaterial() != "" &&
		polygon->getMaterial() != trimesh.material) {
		if (trimesh.material != "") {
		    std::cerr << "Warning: Object '" << mesh.getName()
			<< "' uses multiple materials only 1 material"
			<< " supported for genmesh/sprited3d!"
			<< std::endl;
		} else {
		    trimesh.material = polygon->getMaterial();
		}
	    }
	}
    }
};

void TriangleMesh::convertMesh(const ThingMesh& mesh)
{
    MeshConverter(*this, mesh);
    name = mesh.getName();
}

//-------------------------------------------------------------------

void
TriangleMesh::write(XmlWriter& xml) const
{
    xml.openSection("params");

    // bind the material
    xml.writeTag("material");
    xml.write(getMaterial());

    //vertices
    xml.writeTag("numvt");
    xml.write(getVertexCount());

    VertexIterator v = getVertexIterator();
    while(v.hasNext()) {
        const Vertex& vertex = v.next();
        vertex.position.write(xml, "v");
        xml.writeAttribute("u", vertex.u);
        xml.writeAttribute("v", vertex.v);
    }
    
    // triangles
    xml.writeTag("numtri");
    xml.write(getTriangleCount());

    TriangleIterator t = getTriangleIterator();
    while(t.hasNext()) {
        const Triangle& triangle = t.next();
        xml.writeTag("t");
        xml.writeAttribute("v1", triangle.v[0]);
        xml.writeAttribute("v2", triangle.v[1]);
        xml.writeAttribute("v3", triangle.v[2]);       
    }

    if(flags.smoothing == true)
        xml.writeTag("autonormals");
    if(flags.localshadows == true)
        xml.writeTag("localshadows");
    writeMixMode(xml);

    xml.closeSection("params");
}

} // end of namespace blend2cs

