#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <cstdio>
#include <cmath>
#include <set>
#include <map>

#include "btypes.h"
#include "bfile.h"
#include "bdata.h"
#include "bparse.h"

#include "blendobj.h"
#include "texture.h"
#include "mesh.h"
#include "light.h"
#include "level.h"
#include "object.h"
#include "blendconv.h"

BLFile *blfile;

//std::vector<Structure> structures;
std::string sectorManagement;

char *getString(unsigned int id)
{ // This function assumes that the structure with id id is has a
  // variable named "*next" that has data that points to a character string.
  
  BLStruct *temp=(BLStruct *)id; // ouch -- but hey, it works ;)
  BLData *data;
  int i;
  Variable var;
  
  for(i=0;i<temp->length;i++) {
    data=temp->data[i];
    
    if(std::string("*next")==data->var_name) {  
      // We found the variable, return the pointer to the character string.
      return data->stringPtr;
    }
  }
  return NULL;   // Found nothing appropriate
}

/**
 * Tries to find the DNA structure with the given ID (which actually is a pointer).
 * If followLinks is true then this function will follow the "Link" data structures
 * in the DNA code to their end before returning the ID of the structure.
 */
 
unsigned int FindStructure(unsigned int id, bool followLinks)
{
  BLStruct *temp=(BLStruct *)id;

  if(id==0) {
    std::cerr << "fatal: requested structure with ID 0" << std::endl;
    exit(-1);
  }

  if(std::string(temp->struct_name)==std::string("Link") && followLinks) {
    std::cerr << "warning: following Link when structure with ID " << id << " was requested" << std::endl;
    Variable var;
    var=FindVariable(id,"*next");
    if(var.pointerValue==0uL || var.pointerValue==~0uL) {
      std::cerr << "fatal: requested structure (ID: " << id << ") was a Link with next==NULL" << std::endl;
      exit(-1);
    }
    return FindStructure(var.pointerValue);
  }
  return id;
}

/**
 * This function returns the first DNA structure found in the blender file
 */
unsigned int FindFirstStructure()
{
  return FindStructure((unsigned int)blfile->struct_head);
}

/**
 * This function returns the next DNA structure in the sequence in which they 
 * were parsed from the file. If followLinks is true it will follow DNA
 * "Link" structures like FindStructure() does.
 */
unsigned int FindNextStructure(unsigned int id, bool followLinks)
{
  if(((BLStruct *)id)->next==0) return 0;

  return FindStructure((unsigned int)((BLStruct *)id)->next,followLinks);
}

/**
 * This function return the name of the given structure
 */
std::string GetStructureName(unsigned int id)
{
  return ((BLStruct *)id)->struct_name;
}

/**
 * Convert a variable given as a BLData structure
 * into a Variable.
 */
Variable ConvertVariable(BLData *data)
{
  Variable var;

  var.type=data->type_name;
  var.name=data->var_name;
  var.typeID=data->type;
  var.length=data->length;
  int i;
  switch(data->type) {
  case BFLOAT:
    var.floatValue.resize(var.length);
    for(i=0;i<var.length;i++)
      var.floatValue[i]=data->data.flt[i];
    break;
  case BINT:
    var.intValue.resize(var.length);
    for(i=0;i<var.length;i++)
      var.intValue[i]=BReadINT32((char *)&data->data.intg[i]);
    break;
  case BSHORT:
    var.shortValue.resize(var.length);
    for(i=0;i<var.length;i++)
      var.shortValue[i]=BReadINT16((char *)&data->data.shrt[i]);
    break;
  case BCHR:
    if(var.length==1)
      var.charValue=*data->data.chr;
    else 
      var.stringValue=std::string(data->data.chr).substr(0,var.length);
    break;
  case BPTR:
    var.pointerValue=*data->data.ptr;
    break;
  case BSTRC:
    var.pointerValue=(unsigned long)data->data.strc;
    //    ConvertStruct(data->data.strc);
    break;
  };
  return var;
}

/**
 * Find a structure member by name in a given DNA structure
 * and return the appropriate Variable.
 */
Variable FindVariable(unsigned int id, std::string name) {
  BLStruct *temp=(BLStruct *)id; // ouch -- but hey, it works ;)
  BLData *data;
  int i;
  
  for(i=0;i<temp->length;i++) {
    data=temp->data[i];
    if(name==data->var_name) {
      return ConvertVariable(data);
    }
  }

  std::cerr << "fatal: unable to find structure member by name:" << name << std::endl;
  std::cerr << "members are:" << std::endl;
  for(i=0;i<temp->length;i++)
    std::cerr << temp->data[i]->var_name << std::endl;
  exit(-1);
}


std::set<Mesh, NodeCmp<Mesh> > meshes;
std::set<Light,NodeCmp<Light> > lights;
BObjectContainer instances;

Textures *TextureList;

/**
 * This is the main loop for extracting the data from the DNA structures.
 *
 * @todo briefly explain what it does
 * @todo break it up into smaller functions; it does more than one job
 */
void ExtractData()
{
  unsigned int i;
  std::string currentScene;
  Variable temp;
  int meshID=0,baseID,objID;

  TextureList=new Textures;
  if(TextureList==0) {
    std::cerr << "fatal: unable to allocate texture list object" << std::endl;
    exit(-1);
  }

  i=FindFirstStructure();
  while(i!=0) {
    // Find the Scene structures and follow them.
    if (GetStructureName(i)=="Scene") {
      // Read the current scene, so we can set it on objects that we read.
      // Get the name of the scene.
      temp=FindVariable(i,"id");
      temp=FindVariable(FindStructure(temp.pointerValue),"name[24]");
      currentScene=temp.stringValue.substr(2);
      std::cout << "Parsing Scene " << currentScene << std::endl;

      // Get an id for our linked list of bases.  Each base
      // points to an object.
      temp=FindVariable(i,"base");
      temp=FindVariable(FindStructure(temp.pointerValue),"*first");

      // Loop through the linked Base structures, getting object data.
      while (temp.pointerValue!= (unsigned long) 0 && 
	     temp.pointerValue!= (unsigned long) (~0x0)) {
	BlenderObject  instance;

	// Get the next base identifier.
	baseID=FindStructure(temp.pointerValue);
	if(GetStructureName(baseID)!="Base") {  // It better be a Base
	  std::cerr << "Poo!  referenced structure is not a Base.  It is a "<<
	    GetStructureName(baseID) << std::endl <<
	    "This is a fundamental problem with the program."<< std::endl;
	  exit(1);
	}
	
	// Find the object to which this base points:
	temp=FindVariable(baseID,"*object");
	if (temp.pointerValue==(unsigned long) 0 || 
	    temp.pointerValue== (unsigned long) (~0x0)) {
	  std::cout << "Hey! Invalid pointer in a Base structure.  Report "
		    << " this error, please!."<< std::endl;
	  //  jump to the next base structure
	  temp=FindVariable(baseID,"*next");
	  continue;
	}
	objID=FindStructure(temp.pointerValue);
	
	if(GetStructureName(objID)!="Object") {  // It better be an object
	  std::cerr << "Poo!  referenced structure is not an object.  This" <<
	    " is a fundamental problem with the program."<< std::endl;
	  exit(1);
	}

	// stuff the name of this scene into our set, and set a pointer.
	instance.sceneName=
	  BlenderObject::allMyScenes.insert(currentScene).first;

	std::cout << "found an Object:  ";
	instance.Extract(objID);  // Extract object information.
	std::cout << instance.name<< std::endl;
      
	// ID for structure containing the object's data.
	temp=FindVariable(objID,"*data");
	if (temp.pointerValue)	meshID=FindStructure(temp.pointerValue);
	
	switch(instance.type) {
	case blender_empty:   // object is an empty.
	  std::cout << "it's an empty-> Node" << std::endl;
	  // handle empties here.  Make it a 'node'.
	  instance.data=NULL;
	  break;
	case blender_mesh:   // object is a mesh.
	  {                     // gcc likes the extra little curly braces :)
	    Mesh meshi;
	    // Throw in a dummy check.
	    if (GetStructureName(meshID)!="Mesh") { 
	      std::cerr << "Poo!  referenced structure is not a Mesh. \n"<<
		" This is a fundamental problem with the program."<< std::endl;
	      exit(1);
	    }
	    // Extract its data
	    meshi.Extract(meshID, TextureList);
	    std::cout << "Extracted mesh: " << meshi.name << std::endl;
	    
	    // Insert the mesh into our set if necessary, make a reference
	    instance.data=(BlenderObject *) &(*(meshes.insert(meshi).first));
	    instance.SetMeshFactoryName();
	  }
	  break;
	case blender_light:  // object is a lamp/light.
	  {
	    Light lighti;
	    if (GetStructureName(meshID)!="Lamp") { 
	      std::cerr << "Poo!  referenced structure is not a Lamp. \n"<<
		" This is a fundamental problem with the program."<< std::endl;
	      exit(1);
	    }
	    lighti.Extract(meshID);  // Read light data.
	    
	    std::cout <<"Extracted lamp: "<< lighti.name << std::endl;
	    
	    //Insert the light into our set if necessary, and make a reference
	    instance.data=(BlenderObject *) &(*(lights.insert(lighti).first));
	  }
	  break;
	case blender_camera:  //object is a camera.
	  //  Do some rotation calculations in here.
	  break;
	default:
	  std::cout << "unknown type #" << instance.type << " - rejected" << std::endl;
	  //	  i=FindNextStructure(i,false);
	  temp=FindVariable(baseID,"*next");
	  continue;
	}
        // Store this object, using its id to index it.
        instances.insert(std::make_pair(instance.ID,instance));
	// Now we jump to the next base structure
	temp=FindVariable(baseID,"*next");
      }
    }
    i=FindNextStructure(i,false);
  }

  // Now we go through and match up parent objects to child objects
  // and vice versa.
  BlenderObject *bo;  // heh.
  BObjectContainer::iterator oiter;
  BObjectContainer::iterator parentiter;
  for (oiter=instances.begin(); oiter!=instances.end(); oiter++) {
    bo=& (oiter->second);  // Pointer to this object
    if (bo->parentObjid) {   // If This object has a parent.
      // Resolve parent
      parentiter= instances.find(bo->parentObjid);
      bo->parentObj=  parentiter; // Set this object's parent pointer.
      //This object is it's parent's child.
      parentiter->second.childset.push_back(oiter); 
    }
  }
  
}

int main(int argc, char *argv[])
{
  bool outputLevel;

  if(argc!=5) {
    std::cerr << "usage: " << argv[0] << " [output type] [sector management schemes] [blender file] [output file]" << std::endl;
    std::cerr << std::endl;
    std::cerr << "Valid output types: " << std::endl;
    std::cerr << "  level    output as a CrystalSpace world file" << std::endl;
    std::cerr << "  object   output as a CrystalSpace library file (INCOMPLETE)" << std::endl;
    std::cerr << std::endl;
    std::cerr << "Valid sector management types: " << std::endl;
    std::cerr << "  Scene    each blender scene becomes a CrystalSpace sector" << std::endl;
    std::cerr << "  Layer    each blender layer becomes a CrystalSpace sector" << std::endl;
    std::cerr << "  None     every object is in the \"room\" sector" << std::endl;
    std::cerr << std::endl;
    return -1;
  }

  if((std::string)argv[1]=="level") {
    outputLevel=true;
  } else if((std::string)argv[1]=="object") {
    outputLevel=false;
  } else {
    std::cerr << "usage: " << argv[0] << " {level|object} [blender file] [output file]" << std::endl;
    return -1;
  }

  sectorManagement=argv[2];

  blfile=BLOpenFile(argv[3]);
  if(!blfile) {
    std::cerr << argv[0] << ": unable to open file \"" << argv[3] << "\"" << std::endl;
    return -1;
  }

  BLParseFile(blfile);

  //  BLPrintStructs(blfile);

  ExtractData();

  if(outputLevel) {
    Level *level=new Level;
    level->BuildLevel(&meshes, &lights, &instances);
    std::cout << "Built level"<< std::endl;
    level->Write(argv[4],TextureList);
    std::cout << "Wrote level"<< std::endl;
  } else {
    // TODO: test this code
    Object *object=new Object;
    object->BuildObject(meshes);
    //    TextureList->Write(argv[4]);
    //    object->Write(argv[4],argv[4]);
  }
}
