#include <iostream>
#include <iomanip>
#include <cmath>
#include <map>
#include <utility>

#include "mesh.h"
#include "blendconv.h"

#include "blendobj.h" // for BlenderObject::meshTypNames (which is a BAD HACK)

Mesh::Mesh()
{
  for (int i=0; i<8 ; i++) typesWritten[i]=false;
  nodeType=TMesh;
}

void Mesh::Extract(int meshID, Textures *textures)
{
  Variable temp;
  int i, j;
  int firstVertex;
  int numVertices;
  int firstFace;
  int numFaces;
  int firstUVFace=-1;
  int texID=-1;
  int imageID;
  unsigned int iterator, uvIterator;
  
  // Read BlenderData basics:
  TransformNode::Extract(meshID);
  
  // Extract the number of vertices and faces:
  temp=FindVariable(meshID,"totvert");
  numVertices=temp.intValue[0];
  temp=FindVariable(meshID,"totface");
  numFaces=temp.intValue[0];
  
  // Extract the IDs of the first vertex and face:
  // XXX: The following code assumes that MVert, TFace and MFace data structures are alligned as arrays!
  temp=FindVariable(meshID,"*mvert");
  firstVertex=FindStructure(temp.pointerValue);
  temp=FindVariable(meshID,"*mface");
  firstFace=FindStructure(temp.pointerValue);
  temp=FindVariable(meshID,"*tface");
  if(temp.pointerValue==0) {
    hasUVCoords=false;
    std::cerr << "WARNING: Mesh " << name << " does not have UV coordinates assigend - SKIPPING" << std::endl;
    return;
  } else {
    firstUVFace=FindStructure(temp.pointerValue);
    hasUVCoords=true;
  }


  // Extract the vertex coodinates:
  Vertices.resize(numVertices);
  iterator=firstVertex;
  for(i=0;i<numVertices;i++) {
    //  Read vertex coordinates.
    temp=FindVariable(iterator,"co[3]");
    for(j=0;j<3;j++) {
      Vertices[i].coords.X[j]=temp.floatValue[j];
    }
    // Now read vertex normals
    temp=FindVariable(iterator,"no[3]");
    for(j=0;j<3;j++) 
      Vertices[i].Normals[j]=temp.shortValue[j];

    // Now move to the next vertex;
    iterator=FindNextStructure(iterator);
  }


  // Extract faces and UV coordinates:
  Polys.clear();
  iterator=firstFace;
  uvIterator=firstUVFace;
  hasAnyPortals=false;  // Initialize variable.

  for(i=0;i<numFaces;i++) {
    bool isPortal=false;
    char transparency;
    short faceMode;

    texID=-1;
    
    if(hasUVCoords) {  // If this surface has uv coordinates, grab the image
      std::string thisTexFile,thisTexName;

      temp=FindVariable(uvIterator,"transp"); // Transparency of surface.
      transparency=temp.charValue;

      temp=FindVariable(uvIterator,"mode"); // Transparency of surface.
      faceMode=temp.shortValue[0];

      temp=FindVariable(uvIterator,"*tpage");
      if (!temp.pointerValue==0x0 && temp.pointerValue!=0xffffffff) {
	imageID=FindStructure(temp.pointerValue);
	temp=FindVariable(imageID,"name[160]");   // Name of the image

	thisTexFile=temp.stringValue;      // Name again, Different format
	thisTexName=textures->StripPath(thisTexFile);  
	textures->Add(thisTexFile); // Add this texture to our list of textures
	texID=textures->GetID(thisTexFile);  //Get the ID that we just gave it.

	// Check to see if this is a portal, i.e. does it's name begin 
	// with "Portal"?
	if (thisTexName.find("Portal")==0)  { 
	  hasAnyPortals=isPortal=true;
	} else if (textureFile.empty()) { // set as the default texture for 
	  textureFile=thisTexFile;        // the entire mesh.
	  textureName=thisTexName;
	  this->textures=textures;
	}
      }       
    }


    temp=FindVariable(iterator,"v4");
    if(temp.shortValue[0]==0) { //XXX: Is there a better way to determine that?
      // This face is a triangle
      Polys.resize(Polys.size()+1);
      temp=FindVariable(iterator,"v1");
      Polys[Polys.size()-1].Vertices[0]=temp.shortValue[0];
      temp=FindVariable(iterator,"v2");
      Polys[Polys.size()-1].Vertices[1]=temp.shortValue[0];
      temp=FindVariable(iterator,"v3");
      Polys[Polys.size()-1].Vertices[2]=temp.shortValue[0];
      
      if(hasUVCoords) {
	temp=FindVariable(uvIterator,"uv[4][2]");
	for(j=0;j<3;j++) {
	  Polys[Polys.size()-1].UV[j][0]=temp.floatValue[j*2];
	  Polys[Polys.size()-1].UV[j][1]=temp.floatValue[j*2+1];
	}
	Polys[Polys.size()-1].textureID=texID;  // Assign texture ID to face.
	Polys[Polys.size()-1].isPortal=isPortal; // Mark as a portal or no.
	Polys[Polys.size()-1].transp=(btransparencyType) transparency;
	Polys[Polys.size()-1].faceMode.fm=faceMode;
      } else {
	// TODO: invent some UV coordinates here
      }
    } else {
      // This face is a quad -> split it and store it in two polys
      Polys.resize(Polys.size()+2);
      temp=FindVariable(iterator,"v1");
      Polys[Polys.size()-2].Vertices[0]=temp.shortValue[0];
      temp=FindVariable(iterator,"v2");
      Polys[Polys.size()-2].Vertices[1]=temp.shortValue[0];
      temp=FindVariable(iterator,"v3");
      Polys[Polys.size()-2].Vertices[2]=temp.shortValue[0];
      temp=FindVariable(iterator,"v3");
      Polys[Polys.size()-1].Vertices[0]=temp.shortValue[0];
      temp=FindVariable(iterator,"v4");
      Polys[Polys.size()-1].Vertices[1]=temp.shortValue[0];
      temp=FindVariable(iterator,"v1");
      Polys[Polys.size()-1].Vertices[2]=temp.shortValue[0];
      
      if(hasUVCoords) {
	temp=FindVariable(uvIterator,"uv[4][2]");
	Polys[Polys.size()-2].UV[0][0]=temp.floatValue[0];
	Polys[Polys.size()-2].UV[0][1]=temp.floatValue[1];
	Polys[Polys.size()-2].UV[1][0]=temp.floatValue[2];
	Polys[Polys.size()-2].UV[1][1]=temp.floatValue[3];
	Polys[Polys.size()-2].UV[2][0]=temp.floatValue[4];
	Polys[Polys.size()-2].UV[2][1]=temp.floatValue[5];
	
	Polys[Polys.size()-1].UV[0][0]=temp.floatValue[4];
	Polys[Polys.size()-1].UV[0][1]=temp.floatValue[5];
	Polys[Polys.size()-1].UV[1][0]=temp.floatValue[6];
	Polys[Polys.size()-1].UV[1][1]=temp.floatValue[7];
	Polys[Polys.size()-1].UV[2][0]=temp.floatValue[0];
	Polys[Polys.size()-1].UV[2][1]=temp.floatValue[1];

	// Assign texture ID to faces.
	Polys[Polys.size()-2].textureID=Polys[Polys.size()-1].textureID=texID;
	// Eventually mark poly as a portal
	Polys[Polys.size()-2].isPortal=isPortal;
	Polys[Polys.size()-1].isPortal=isPortal;
	Polys[Polys.size()-2].transp=(btransparencyType) transparency;
	Polys[Polys.size()-1].transp=(btransparencyType) transparency;
	Polys[Polys.size()-2].faceMode.fm=faceMode;
	Polys[Polys.size()-1].faceMode.fm=faceMode;
      } else {
	// TODO: invent some UV coordinates here
      }
    }
    iterator=FindNextStructure(iterator);
    if(hasUVCoords)
      uvIterator=FindNextStructure(uvIterator);
  }

  std::cout << "extracted " << Vertices.size() << " vertices and " << Polys.size() << " polygons" << std::endl;

  //  ApplyTransformation();  // Swap vertices.
  ReduceData();

  std::cout << "reduced to " << Vertices.size() << " vertices and " << Polys.size() << " polygons" << std::endl;
}

bool Mesh::HasPortals()
{  // function returns true if this mesh has any portals.
  return hasAnyPortals;
}

void Mesh::WriteFactory(std::ofstream &output,std::string &factName, int typeOfMesh)
{
  if(!hasUVCoords) return;

  //Check if We've already written out this type of mesh to a factory.
  if (typesWritten[typeOfMesh]) return; 
  
  // Soon we will have written out this type of mesh for this mesh data.
  typesWritten[typeOfMesh]=true;  

  // The name of the mesh factory is the name of the blender mesh.
  output << "    <meshfact name=\"" << factName << "\">" << std::endl;
  // Print the type of the plugin that you want to use.
  output << "      <plugin>crystalspace.mesh.loader.factory." << 
    BlenderObject::meshTypeNames[typeOfMesh] << "</plugin>"<< std::endl;

  Write(output,typeOfMesh);  // Write the actual data.
  output << "    </meshfact>" << std::endl;
}

void Mesh::PropertyMapCopy(PropertyMap &pMap)
{
  propMap=pMap; //Set this mesh's property map to the supplied value.
}

void Mesh::Write(std::ofstream &output,int typeOfMesh)
{
  unsigned int i;
  std::string sectorN;

  if(!hasUVCoords) return;

  output << "      <params>" << std::endl;

  // Output default texture
  output << "        <material>" << textureName << "</material>" << std::endl;

  if (BlenderObject::meshTypeNames[typeOfMesh]=="thing") {  // Write output for a "thing"

    for(i=0;i<Vertices.size();i++) 
      output << "        <v " << MakeCoordinateString(Vertices[i].coords) << "/>" 
	     << std::endl;

    for(i=0;i<Polys.size();i++) Polys[i].WriteThing(output, propMap,textures);

  } else if (BlenderObject::meshTypeNames[typeOfMesh]=="genmesh") {  // It's a Genmesh 
    output << "        <numvt>"<< Vertices.size() << "</numvt>"<< std::endl;
    output << "        <numtri>"<< Polys.size() << "</numtri>"<< std::endl;

    WriteVertices(output);  // Print out vertices.
    WriteTriangles(output);  // Print out triangles
 
    // We may have to specify normals. who knows
    output << "        <autonormals />" << std::endl; 
    
  } else if (BlenderObject::meshTypeNames[typeOfMesh]=="sprite.3D") {  // It's a 3D sprite
    output << "    <frame name=\"0\">" << std::endl;
    WriteVertices(output);
    output << "    </frame>" << std::endl;
    output << "    <action name=\"default\">" << std::endl;
    output << "      <f name=\"0\" delay=\"1000\" />" << std::endl;
    output << "    </action>" << std::endl;
    WriteTriangles(output);
  }

  
  output << "      </params>" << std::endl;
}


void Mesh::WriteVertices(std::ofstream &output)
{  // This routine prints out triangles for genmeshes and 3d sprites.
  // Although 3d sprites will probably use something else in the future,
  // as they move.
  for(unsigned int i=0;i<Vertices.size();i++) {
    double *uv;
    // Find a face that contains this vertex to get the uv coords
    GetUVCoords(i,&uv);
    
    output << "        <v " << MakeCoordinateString(Vertices[i].coords) << 
      "u=\"" << uv[0] << "\" v=\"" << -uv[1] <<   "\" />" << std::endl;
  }
}

void Mesh::WriteTriangles(std::ofstream &output)
{
  // This routine prints out triangles for genmeshes and 3d sprites.
  for(int i=0;i<Polys.size();i++) {   // Now print out the triangles
    output << "        <t ";
    for(int j=2;j>=0;j--) 
      output << "v"<< 2-j+1 << "=\"" << Polys[i].Vertices[j] << "\" ";
    output << "/>" << std::endl;
  }
}

void Mesh::GetUVCoords(int vertexNumber,double **uv)
{  // this routine simply returns the UV coordinates for a given vertex.
  int j,k;
  
  for (j=0; j< Polys.size(); j++) { //look at every polygon 
    for (k=0;k<3;k++)  {   // and look at every vertex of every polygon
      // To see if it is the vertex that we want.
      if (Polys[j].Vertices[k]==vertexNumber) { 
	// From that polygon's vertex we can get the UV mapping.
	*uv=Polys[j].UV[k];
	return ;
      }
    }
  }
  
  // Oooh.  Pooper.  there were no 
  std::cerr << " Ooh.  You got to the part of the program where the " <<
    "programmer didn't know what to do." << std::endl;
  exit(1);
}


void Mesh::ReduceData()
{
  unsigned int i, j;
  int k, l;
  int numMatches;
  bool isEqual, isValid;
  std::vector<int> vertexMap;
  std::vector<Vertex> newVertices;
  std::vector<Poly> newPolys;

  // try to eliminate duplicate vertices:
  
  vertexMap.resize(Vertices.size());

  for(i=0;i<Vertices.size();i++) {
    isEqual=false;
    for(j=0;j<newVertices.size();j++) {
      if((float)Vertices[i].coords.X[0]==(float)newVertices[j].coords.X[0] &&
	 (float)Vertices[i].coords.X[1]==(float)newVertices[j].coords.X[1] &&
	 (float)Vertices[i].coords.X[2]==(float)newVertices[j].coords.X[2]) { 
	isEqual=true;
	vertexMap[i]=j;
	break;
      }
    }
    if(!isEqual) {
      newVertices.push_back(Vertices[i]);
      vertexMap[i]=newVertices.size()-1;
    }
  }

  Vertices=newVertices;

  for(i=0;i<Polys.size();i++) {
    for(j=0;j<3;j++) {
      Polys[i].Vertices[j]=vertexMap[Polys[i].Vertices[j]];
    }
  }

  // try to eliminate invalid/duplicate polygons:
  for(i=0;i<Polys.size();i++) {
    // HACK: replace certain UV coordinates as they seem to be special:
    //    if(Polys[i].UV[0][0]==0 && Polys[i].UV[0][1]==-1 &&
    //       Polys[i].UV[1][0]==0 && Polys[i].UV[1][1]==0 &&
    //       Polys[i].UV[2][0]==0 && Polys[i].UV[2][1]==0) {
    //      Polys[i].UV[2][0]=-1;
    //    }
    if(Polys[i].Vertices[0]==Polys[i].Vertices[1] ||
       Polys[i].Vertices[0]==Polys[i].Vertices[2] ||
       Polys[i].Vertices[1]==Polys[i].Vertices[2]) {
      std::cerr << "WARNING: found polygon with only two distinct vertices -- dropping" << std::endl;
      isValid=false;
    } else {
      isValid=true;
    }
    if(isValid) {
      isEqual=false;
      for(j=0;j<newPolys.size();j++) {
	numMatches=0;
	for(k=0;k<3;k++) {
	  for(l=0;l<3;l++) {
	    if(Polys[i].Vertices[k]==Polys[j].Vertices[l]) {
	      numMatches++;
	    }
	  }
	}
	if(numMatches==3) {
	  isEqual=true;
	  break;
	} else if(numMatches>3) {
	  std::cerr << "INTERNAL ERROR: numMatches > 3" << std::endl;
	  exit(-1);
	}
      }
      if(!isEqual) {
	newPolys.push_back(Polys[i]);
      }
    }
  }
  Polys=newPolys;
}
