// $Id: ExternalExec.h 72 2003-10-24 23:58:51Z felfert $
//

#if !defined(AFX_EXTERNALEXEC_H__A88D724E_B460_4B18_B8FB_CE59506970DA__INCLUDED_)
#define AFX_EXTERNALEXEC_H__A88D724E_B460_4B18_B8FB_CE59506970DA__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <deque>

class Process;

#ifdef __cplusplus
extern "C" {
#endif

typedef BOOL (__cdecl *runCallBack)(LPVOID, BOOL, const char *);
typedef void (__cdecl *waitCallBack)(LPVOID);

#ifdef __cplusplus
}
#endif

typedef enum pipeTypes {
	PT_STDIN	= 0,
	PT_STDOUT	= 1,
	PT_STDERR	= 2,
} TpipeTypes;

typedef enum errorTypes {
	DUPLICATEHANDLE_FAILED	= 1,
	CREATEPIPE_FAILED		= 2,
	PROCESSLIST_EMPTY		= 3,
	CREATETHREAD_FAILED		= 4,
	CREATEPROCESS_FAILED	= 5,
	INVALID_STDINHANDLE		= 6,
} TerrorTypes;

/**
* Helper class for executing several external processes in a chain on Windows.
* Just like a Unix shell, it connects stdin of one process to stdout of the next process.
* The last process is connected to stdout of the parent. The first process can optionally
* be connected to the parent also. If this is the case, the parent can use the Write() member
* to supply input to the first process. When it is finished, it must call the Close() member to
* signal end of input to the first process. All processes stderr descriptors are connected
* to the parent. The parent can optionally specify a callback function which get's called
* whenever a line is available on any childs stderr or on stdout of the last process.
*
* @Author: Fritz Elfert <felfert@to.com>
*/
class ExternalExec
{
public:
	/**
	* Add a process to the process-chain
	* @Parameters:
	*	executable:	The path to the program to be run as child
	*	params:		The parameters to this program. (On execution, the basename of the
	*				executable gets automatically inserted at the beginning of the command line.
	*	wdir:		The working directory for this executable.
	*	showWindow:	The showWindow parameter of CreateProcess()
	*
	* @Returns:
	*	true on success, false otherwise.
	*/
	bool AddProcess(const char *executable, const char *params = NULL, const char *wdir = NULL, int showWindow = SW_SHOW);

	/**
	* Start process chain.
	* @Parameters:
	*	cbParam:	A user-defined pointer which gets used as first parameter of any callback.
	*	needStdin:	If true, a pipe to stdin of the first process is created. Must use Close() to finish output.
	*
	* @Returns:
	*	true on success, false otherwise. (Note: returns immediately after all processes have been created)
	*/
	bool Run(LPVOID cbParam = NULL, bool needStdin = false);

	/**
	* Specify a callback function which gets called whenever any output is available.
	* @Parameters:
	*	pointer to a user-defined function.
	*
	*	This function gets called with 3 parameters:
	*		1st:	The user-defined pointer, specified as cbParam when calling Run()
	*		2nd:	A boolean, indicating stderr. TRUE, if this callback is triggered by a stderr output from any process,
	*				FALSE, if it is triggered by stdout output from the last process.
	*		3rd:	The asciiz string, containing the text which triggered the callback.
	*
	*   If it returns TRUE, the data-punp continues. If it returns FALSE, the childs handles are closed which usually
	*   results in termination of the child process.
	*/
	void SetCallBack(runCallBack);

	/**
	* Specify a callback for waiting purposes.
	* @Parameters:
	*	pointer to a user defined function.
	*
	*	This function gets called when in either the destructor or the ExitCode() member is called and at least one
	*	child process is still running, or when the Write() member is called and the first process does not read from its
	*	stdin.
	*
	*	This function gets called with 1 parameter: The user-defined pointer, specified as cbParam when calling Run()
	*/
	void SetWaitCallBack(waitCallBack);

	/**
	* Write data to stdin of the first child process. This function does not return until either the data has been written or
	* an error happened.
	* @Parameters:
	*	asciiz string to be written.
	*
	* @Returns:
	*	true on success, false otherwise.
	*/
	bool Write(const char *);

	/**
	* Close stdin of the first process.
	*
	* @Returns:
	*	true on success, false otherwise.
	*/
	bool Close();

	/**
	* Check for running child processes.
	*
	* @Returns:
	*	true, if any child is running, false otherwise.
	*/
	bool Running();

	/**
	* retrieve the exit code of the last child process.
	* @Parameters:
	*	flush: If true, flush output pipe.
	*
	* @Returns:
	*	The exit code of the last process in the child process chain.
	*	If the process is still running, never started or the process chain is empty: returns STILL_ACTIVE
	*/
	DWORD ExitCode(bool flush = true);

	/**
	* retrieve the last error code.
	*
	* @Returns:
	*	A code, desribing the type of the last error.
	*/
	DWORD LastError();

	/**
	* Kill the running childs. USE WITH EXTREME CAUTION!
	* Closes all file descriptors and waits 1 sec for each child to terminate. If it does not
	* terminate within this time, first tries to send Ctrl-C, Ctrl-Break and if this still did not
	* terminate the child, performs TerminateProcess() on the child process. TerminateProcess()
	* can lead to LOADS of inconsistencies! USE WITH EXTREME CAUTION!
	*/
	void Kill();

	/**
	* Deletes process list and resets callback pointers.
	* If child processes are running, does nothing.
	*
	* @Returns:
	*   true on success, false on failure (processes still running)
	*/
	bool Clear();

	/**
	* Sets fixed buffersize of internal IPC pump. (0 = line-buffering)
	*/
	void SetIpcBuffer(int);

	/**
	* Standard constructor
	*/
	ExternalExec();

	/**
	* Standard destructor. The destructor does not return until all child processes have terminated.
	*/
	virtual ~ExternalExec();

private:
	std::string				m_strStdinBuf;
	std::deque<Process*>	m_vProcessList;
	HANDLE					m_hStdHandle[6];
	runCallBack				m_pfCallBack;
	waitCallBack			m_pfWaitCallBack;
	LPVOID					m_lpvCallbackParam;
	bool					m_bCloseStdin;
	CRITICAL_SECTION		m_sHandleModify;
	CRITICAL_SECTION		m_sProcessCountAccess;
	HANDLE					m_hStdinWriteEvent;
	HANDLE					m_hStdinWrittenEvent;
	HANDLE					m_hStdinThread;
	HANDLE					m_hStdoutThread;
	HANDLE					m_hStderrThread;
	DWORD					m_dwLastError;
	long					*m_piProcessCount;
	int						m_iIpcBuffer;

	bool createStdPipe(TpipeTypes);
	void cleanup();
	void outPump(TpipeTypes);
	void inPump();

	static void stdinPump(LPVOID);
	static void stdoutPump(LPVOID);
	static void stderrPump(LPVOID);
};

#endif // !defined(AFX_EXTERNALEXEC_H__A88D724E_B460_4B18_B8FB_CE59506970DA__INCLUDED_)
