/********************************************************************
**                                                                 **
**  Blender File          author: celer                            **
**  Library               email: celer@scrypt.net                  **
**                        Copywrite (C) 2000 by David Tyree.       **
**                        All rights reserved.                     **
*********************************************************************
* This library is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU Library General Public       *
* License as published by the Free Software Foundation; either      *
* version 2 of the License, or (at your option) any later version.  *
*                                                                   *
* This library is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *
* Library General Public License for more details.                  *
*                                                                   *
* You should have received a copy of the GNU Library General Public *
* License along with this library; if not, write to the Free        *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.*
*********************************************************************
$Id: bdata.h 20 2003-05-20 19:53:33Z kludge $
********************************************************************/ 
#ifndef BDATA_H
#define BDATA_H

BEGIN_C_DECLS

/*
 These functions deal with the blender data after it has been parsed.
 Blender stores data in it's files in c structures. The structure definition
 is written to the output file along with the actual data

 After parsing the output file we see this data as a structer with some number
 of members. When a member is a pointer to another structure we try to resolve it
 if we were unable to resolve it, we return a pointer value of 0xFFFFFFFF.

 Look at the default.out in the blend directory to get an idea of how the output
 looks.

*/


/* Basic C data structures */
enum { BFLOAT=0, BCHR,BINT, BSHORT, BPTR, BSTRC } BTypes;

struct _BLStruct ;

/* Structure for a single variable in the a structure 
	*You should try to access this function through the
	provided functions..*
*/
typedef struct {
	char *var_name;			/* Variable name */
	char *type_name;		/* Type name */
	
	/* Union of possible data types */
	union {
			float *flt;
			int *intg;
			short *shrt;
			uint32 *ptr;
			char *chr;
			struct _BLStruct *strc;	
	} data;
  char  *stringPtr;
  //	uint16 type;	/* Value of the one of the enums from BTypes */
        uint8 type;
        uint16 length; 	/* Length of the data structure.. incase of arrays */
} __attribute__ ((packed)) BLData;


/* Structure defining a blender structure */
typedef struct _BLStruct {
	char *struct_name;	/* Structure name */
	void *ptr;		/* File pointer.. DONT USE THIS FOR ANYTHING!!  */
	BLData **data;		/* Pointer to an array of memeber data */
	uint16 length;		/* Length of the array in members */
	struct _BLStruct *next;	/* Next structure */
	struct _BLStruct *prev;	/* Prev structure */
} __attribute__ ((packed)) BLStruct;

/* BLFindStruct 
 Given the head of a the linked list find the structure
 matching the given name.
	head - head of the structure list
	name - name of the structure to find 

	returns NULL upon failure to find a structure
*/
BLStruct *BLFindStruct(BLStruct *head, const char *name);

/* BLFindData 
 Given the structure find the data element specified
 by the given path. For instance "id.newid", would
 return the newid data element in the id struct under 
 the specified struct.
 	strc - structure to look in
	path - path to the specified data 

	returns NULL upon failure to find a data member
*/
BLData *BLFindData(BLStruct *strc, const char *path);


/* BLPrintData 
 Print out the given data element, prefixing it with
 *level* tabs. This routine is used to recursively
 print out the data in a structure.
	data - data to print out
	level - the number of tabs to print
*/
void BLPrintData(BLData *data, uint16 level);


/* BLPrintStruct
 Print out all the structures in the BLFile
	file -- file structure to print out
*/
void BLPrintStructs(BLFile *file);

/* BLPrintStruct 
 Print out the specified BLStruct prefixed
 with *level* many tabs.
	strc - the BLStruct to print out
	level - the number of tabs to print out
*/
void BLPrintStruct(BLStruct *strc, uint16 level);

/* BLFreeStructs
 Free all the structs and
 data memebers

*/
void BLFreeStructs(BLStruct *head);

/***************************Internal Functions**************************/


/* BLNewData
 Allocate and initialize a new BLData struct.
 	return new BLData struct.
*/
BLData *BLNewData(void);

/* BLInitData*
 Initialize a new BLData struct to the specified name, type, and length
	var - variable name
	type - variable type
	XXXX - pointer to data *This is not copied*
	len - the length of the data in question 
*/
BLData *BLInitDataFloat(char *var, char *type, float *flt, uint16 len);
BLData *BLInitDataChar(char *var, char *type, char *chr, uint16 len);
BLData *BLInitDataInt(char *var, char *type, uint32 *intg, uint16 len);
BLData *BLInitDataShort(char *var, char *type, short *shrt, uint16 len);
BLData *BLInitDataPointer(char *var, char *type, void *ptr, uint16 len);
BLData *BLInitDataStruct(char *var, char *type, BLStruct *ptr, uint16 len);

/* BLNewStruct
 Allocate and initialize a new BLStruct struct.

	return new BLStruct
*/
BLStruct *BLNewStruct(void);

/* BLInitStruct 
 Allocate and initialize a new BLStruct struct
 	name - the name of the structure
	ptr - the FILE pointer (INVALID)
	length - number of BLData members
*/
BLStruct *BLInitStruct(char *name, void *ptr, uint16 length);

/* BLFreeData
 Free the specified BLData struct 

*/
void BLFreeData(BLData *data);

/* BLFreeStruct
	Free the specified BLStruct
*/
void BLFreeStruct(BLStruct *data);


END_C_DECLS


#endif
